import asyncio
import argparse
from pathlib import Path
from .manager import PipelineManager
from .utils import *


def setup_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('-p', '--pipeline', type=str, required=True, help='specify the pipeline to run')
    parser.add_argument('-i', '--input', type=str, help='specify input data')
    parser.add_argument('-o', '--output', type=str, help='specify output data')
    parser.add_argument('-m', '--mode', type=str, default='async', choices=['async', 'mp', 'seq'], help='specify the run mode')
    parser.add_argument('--split', type=int, help='split the data into parts to run')
    parser.add_argument('--png', action='store_true', help='export graph as png')
    parser.add_argument('--test', action='store_true', help='run test')

    args, _ = parser.parse_known_args()
    return args

def main():
    args = setup_args()
    pipefile = Path(args.pipeline)

    pm = PipelineManager(run_mode=args.mode)
    pipe = pm.add_pipe(pipefile.stem, pipefile=pipefile)

    if args.input:
        data = jload(args.input)
        r = pipe.run(data, split=args.split, save_perf=args.png)

        if args.output:
            if type(r) is tuple: jdump(r[0], args.output)
            elif type(r) is list: jdump([i for i,_ in r], args.output)
    elif args.png:
        pipe.to_png(f'{pipefile.stem}.png')




from . import *
spec = importlib.util.find_spec('vllm')

if spec:
    from vllm import LLM, SamplingParams

    model_path = os.getenv('VLLM_MODEL')
    max_model_len = int(os.getenv('VLLM_MAX_MODEL_LEN', 256))
    max_tokens = int(os.getenv('VLLM_MAX_TOKENS', 256))
    batch_size = int(os.getenv('VLLM_BATCH_SIZE', 128))
    pp = int(os.getenv('VLLM_PP', 1))
    tp = int(os.getenv('VLLM_TP', 1))

    llm = LLM(model=model_path,
            gpu_memory_utilization=.95,
            max_model_len=max_model_len,
            pipeline_parallel_size=pp,
            tensor_parallel_size=tp)
    sampling_params = SamplingParams(temperature=0.7,
                                    max_tokens=max_tokens,
                                    min_p=0.15,
                                    top_p=0.85)

batch_wait_time = 1
batch_queue = []
batch_event = asyncio.Event()

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    raise NotImplementedError

def batch_completion(batch_data):
    prompts = [text for text,_ in batch_data]
    outputs = llm.generate(prompts, sampling_params)
    result = [o.outputs[0].text for o in outputs]
    return result

async def llm_batch_processor():
    global batch_queue
    while True:
        await batch_event.wait()

        if batch_queue:
            batch = batch_queue[:batch_size]
            batch_queue = batch_queue[len(batch):]
            result = await asyncio.to_thread(batch_completion, batch)

            for (_, future), res in zip(batch, result):
                future.set_result(res)

        if not batch_queue:
            batch_event.clear()

async def async_completion(text):
    global batch_queue

    loop = asyncio.get_event_loop()
    future = loop.create_future()
    batch_queue.append((text, future))

    if len(batch_queue) >= batch_size:
        batch_event.set()
    else:
        await asyncio.sleep(batch_wait_time)
        batch_event.set()

    return await future


import sys
import json
import shutil
import hashlib
import collections
from pathlib import Path, PosixPath

def check_cmd_exist(command):
    return shutil.which(command) is not None

def importpath(path):
    if type(path) is not PosixPath:
        strpath = str(path)
        if not strpath.startswith("/"):
            parent_path = Path(sys._getframe().f_globals.get("__file__", ".")).parent
            path = parent_path / path
        else:
            path = Path(path)

    try:
        sys.path.insert(0, str(path.parent))
        module = __import__(path.stem)
    finally:
        sys.path.pop(0)
    return module

def get_ordered_task(tasks):
    def get_dependencies(task_id, task_info):
        dependencies = []
        for key, value in task_info["inputs"].items():
            if isinstance(value, list):
                for item in value:
                    dep_task, _ = item.split("-")
                    dependencies.append(dep_task)
        return dependencies

    # Step 1: Build dependency graph and indegree count
    graph = collections.defaultdict(list)
    indegree = collections.defaultdict(int)

    for task_id, task_info in tasks.items():
        dependencies = get_dependencies(task_id, task_info)
        for dep in dependencies:
            graph[dep].append(task_id)
            indegree[task_id] += 1

    # Step 2: Topological sort using Kahn's algorithm
    execution_order = []
    queue = collections.deque([task_id for task_id in tasks if indegree[task_id] == 0])

    while queue:
        current = queue.popleft()
        execution_order.append(current)

        for neighbor in graph[current]:
            indegree[neighbor] -= 1
            if indegree[neighbor] == 0:
                queue.append(neighbor)

    return execution_order

def jload(inp):
    with open(inp, 'r', encoding='utf-8') as f:
        return json.load(f)

def jdump(obj, out):
    with open(out, 'w', encoding='utf-8') as f:
        if isinstance(obj, (dict, list)):
            json.dump(obj, f, indent=4, ensure_ascii=False)
        elif isinstance(obj, str):
            f.write(obj)
        else:
            raise ValueError(f"Unexpected type: {type(obj)}")

def calc_sha256(file_path):
    hash_sha256 = hashlib.sha256()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()



from . import *
spec = importlib.util.find_spec('lmdeploy')

if spec:
    from lmdeploy import pipeline, GenerationConfig, TurbomindEngineConfig, ChatTemplateConfig

    model_path = os.getenv('LMDEPLOY_MODEL')
    chat_temp_name = os.getenv('LMDEPLOY_CHAT_TEMP', None)
    session_len = int(os.getenv('LMDEPLOY_SESSION_LEN', 256))
    max_new_tokens = int(os.getenv('LMDEPLOY_MAX_NEW_TOKENS', 256))
    batch_size = int(os.getenv('LMDEPLOY_BATCH_SIZE', 128))
    pp = int(os.getenv('LMDEPLOY_PP', 1))
    tp = int(os.getenv('LMDEPLOY_TP', 1))

    backend_config = TurbomindEngineConfig(
                        session_len=session_len,
                        pp=pp,
                        tp=tp)
    gen_config = GenerationConfig(
                        do_sample=True,
                        # top_p=0.95,
                        # temperature=0.7,
                        max_new_tokens=max_new_tokens)
    chat_temp = ChatTemplateConfig(model_name=chat_temp_name) if chat_temp_name else None
    llm = pipeline(model_path, backend_config=backend_config, chat_template_config=chat_temp)

batch_wait_time = 1
batch_queue = []
batch_event = asyncio.Event()

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    raise NotImplementedError

def batch_completion(batch_data):
    prompts = []
    for text, _ in batch_data:
        if (t := type(text)) is str:
            prompts.append(text)
        elif t is list:
            msg = []
            for role, c in zip(['user', 'assistant']*len(text), text):
                msg.append({
                    "role": role,
                    "content": c,
                })
            prompts.append(msg)

    outputs = llm(prompts, gen_config=gen_config)
    result = [o.text for o in outputs]
    return result

async def llm_batch_processor():
    global batch_queue
    while True:
        await batch_event.wait()

        if batch_queue:
            batch = batch_queue[:batch_size]
            batch_queue = batch_queue[len(batch):]
            result = await asyncio.to_thread(batch_completion, batch)

            for (_, future), res in zip(batch, result):
                future.set_result(res)

        if not batch_queue:
            batch_event.clear()

async def async_completion(text):
    global batch_queue

    loop = asyncio.get_event_loop()
    future = loop.create_future()
    batch_queue.append((text, future))

    if len(batch_queue) >= batch_size:
        batch_event.set()
    else:
        await asyncio.sleep(batch_wait_time)
        batch_event.set()

    return await future


from . import *
import torch
from transformers import pipeline

model_path = os.getenv('MODEL_PATH')
max_new_tokens = int(os.getenv('MAX_TOKENS', 256))

pipe = pipeline(
    "text-generation", 
    model=model_path, 
    torch_dtype=torch.bfloat16, 
    device_map="auto"
)

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    if (t := type(text)) is str:
        msg = [
            {
                "role": "user",
                "content": text,
            }
        ]
    elif t is list:
        msg = []
        for role, c in zip(['user', 'assistant']*len(text), text):
            msg.append({
                "role": role,
                "content": c,
            })

    resp = pipe(
        msg,
        do_sample=True,
        max_new_tokens=max_new_tokens,
        pad_token_id = pipe.tokenizer.eos_token_id
    )

    return resp[0]["generated_text"][-1]['content']

async def async_completion(text):
    async with llm_sem:
        resp = completion(text)

    return resp



import os
from mlx_lm import load, generate
from . import *

model, tokenizer = load(os.getenv('MLX_MODEL'))
max_tokens = int(os.getenv('MLX_MAX_TOKENS', 256))

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    messages = [{"role": "user", "content": text}]
    prompt = tokenizer.apply_chat_template(
        messages, tokenize=False, add_generation_prompt=True
    )
    text = generate(model, tokenizer, prompt=prompt, verbose=False, max_tokens=max_tokens)

    return text

async def async_completion(text):
    return completion(text)



import os, sys
sys.path.append('/Users/mkk/workspace/git_repos/mlx_parallm')
from mlx_parallm.utils import load, batch_generate
from . import *

model, tokenizer = load(os.getenv('MLX_MODEL'))
max_tokens = int(os.getenv('MLX_MAX_TOKENS', 256))
batch_size = int(os.getenv('MLX_BATCH_SIZE', 128))

batch_wait_time = 1
batch_queue = []
batch_event = asyncio.Event()

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    raise NotImplementedError

def batch_completion(batch_data):
    print(f"Processed batch, len: {len(batch_data)}")
    prompts = [text for text,_ in batch_data]
    result = batch_generate(model, tokenizer, prompts=prompts, verbose=True, format_prompts=True, max_tokens=max_tokens, temp=0.7)
    return result

async def llm_batch_processor():
    global batch_queue
    while True:
        await batch_event.wait()
        print("Batch processor triggered.")

        if batch_queue:
            batch = batch_queue[:batch_size]
            batch_queue = batch_queue[len(batch):]
            result = await asyncio.to_thread(batch_completion, batch)

            for (_, future), res in zip(batch, result):
                future.set_result(res)

        if not batch_queue:
            print("Batch queue is empty. Resetting event.")
            batch_event.clear()

async def async_completion(text):
    global batch_queue

    loop = asyncio.get_event_loop()
    future = loop.create_future()
    batch_queue.append((text, future))
    print("Task added to batch queue.")

    if len(batch_queue) >= batch_size:
        print("Batch size reached. Triggering batch processing.")
        batch_event.set()
    else:
        print(f"Task waiting for batch timeout or size threshold.")
        await asyncio.sleep(batch_wait_time)
        batch_event.set()

    return await future


from . import *
import json
import httpx
from fastapi import HTTPException

pulse_model = os.getenv('PULSE_MODEL')
pulse_token = os.getenv('PULSE_TOKEN')
pulse_url = os.getenv('PULSE_URL')
max_tokens = int(os.getenv('PULSE_MAX_TOKENS', 2048))
retry_count = int(os.getenv('PULSE_RETRY_COUNT', 5))
repetition_penalty = float(os.getenv('PULSE_REPETITION_PENALTY', 1))
temperature = float(os.getenv('PULSE_TEMPERATURE', 0.7))
top_p = float(os.getenv('PULSE_TOP_P', 0.1))
top_k = int(os.getenv('PULSE_TOP_K', 10))

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    payload = json.dumps({
        "action": "To user",
        "parent_messages": [
            {
                "action": "From user",
                "content": text
            }
        ],
        "gen_kwargs": {
            "model": pulse_model,
            "num_return_sequences": 1,
        }
    })
    headers = {
        'accept': 'application/json',
        'Authorization': f'Bearer {pulse_token}',
        'Content-Type': 'application/json'
    }

    response = requests.request("POST", pulse_url, headers=headers, data=payload, verify=False)
    if response.status_code != 200:
        raise Exception(f"Failed to generate completion: {response.text}")

    return response.json()['messages'][0]['content']['parts'][0]

async def async_completion(text):
    messages = [
        {
            'action': 'From user', # 'To user'
            'content': text,
        }
    ]

    async with llm_sem:
        input_data = {
            "action": "To user",
            "parent_messages": messages,
            "gen_kwargs": {
                "model": pulse_model,
                "num_return_sequences": 1,
                "temperature": temperature,
                "top_p": top_p,
                "top_k": top_k,
                "max_tokens": max_tokens,
                "repetition_penalty": repetition_penalty,
            }
        }

        return await _get_completion(
            input_data=input_data,
            retry=retry_count
        )

async def _get_completion(input_data: dict, retry: int = 5):
    try:
        async with httpx.AsyncClient(verify=False) as client:
            async with client.stream(
                "POST",
                pulse_url,
                json=input_data,
                headers={
                    "Authorization": f"Bearer {pulse_token}",
                },
                timeout=600,
            ) as stream:
                # error
                if stream.status_code != 200:
                    error_detail = ""
                    try:
                        error_detail = (await stream.aread()).decode("utf8")
                        error_detail = str(json.loads(error_detail))
                    except:
                        pass

                    raise HTTPException(status_code=500, detail=error_detail)

                content = (await stream.aread()).decode("utf8")
                content = json.loads(content)
                content = "".join(content['messages'][0]['content']['parts'])
                return content
    except Exception as e:
        # logger.error(json.dumps({
        #     "retry": retry,
        #     "input": input_data,
        #     "error": traceback.format_exc()
        # }, ensure_ascii=False, indent=4))
        # 重试次数到达极限
        if retry == 0: raise e

        return await _get_completion(
            input_data=input_data,
            retry=retry-1,
        )



import os
import tqdm
import time
import json
import uuid
import asyncio
import datetime
import collections
from tqdm.asyncio import tqdm_asyncio
from .log import log, log_dir
from .pipetree import PipeTree
from .utils import *

class Pipeline:
    def __init__(self, llm_backend, rag_backend, prompt_manager, pipeconf=None, pipefile=None, run_mode='async'):
        self.llm_backend = llm_backend
        self.rag_backend = rag_backend
        self.run_mode = run_mode
        self.pipefile = pipefile
        self.name = f'pipeline-{datetime.datetime.now().strftime("%m/%d/%Y_%H:%M:%S")}'
        self.hash = ''
        if pipeconf or pipefile:
            self.pipetree = PipeTree(llm_backend, rag_backend, prompt_manager, pipeconf=pipeconf, pipefile=pipefile, run_mode=run_mode)
            self.name = self.pipetree.name
            if pipefile: self.hash = calc_sha256(pipefile)

    def gen_info(self, data, start_t, save_perf=False):
        pipe_manager = self.pipetree.pipe_manager

        info = {
            'perf': self.pipetree.perf,
            'exec_path': [n[1] for n in self.pipetree.perf],
            'detail': {},
            'total_time': time.time()-start_t,
            'mermaid': {},
        }

        for k in sorted(pipe_manager, key=lambda k: pipe_manager[k].time or -1):
            info['detail'][k] = {
                # 'run_time': list(pipe_manager[k].run_time),
                'avg_time': pipe_manager[k].time,
            }

        if save_perf and 'error_msg' not in data:
            info['mermaid']['pipe'] = self.pipetree.tree2mermaid(info)
            info['mermaid']['perf'] = self.pipetree.perf2mermaid()
            fname = datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
            if check_cmd_exist('mmdc'):
                pipe_img = str(log_dir / f'{fname}_pipe.png')
                tmp_file = f'/tmp/{uuid.uuid4()}'
                with open(tmp_file, 'w') as f: f.write(info["mermaid"]["pipe"])
                log.debug(f'Save pipeline mermaid in: {tmp_file}')
                os.popen(f'mmdc -i {tmp_file} -o {pipe_img} -s 3 >/dev/null 2>&1')
                log.debug(f'save {pipe_img}')
                perf_img = str(log_dir / f"{fname}_perf.png")
                os.popen(f'echo "{info["mermaid"]["perf"]}" | mmdc -i - -o {perf_img} -s 3 >/dev/null 2>&1')
                log.debug(f'save {perf_img}')
                # md_pipe = f"![pipe_img]({pipe_img.split('/')[1]})"
                # md_perf = f"![perf_img]({perf_img.split('/')[1]})"
            else:
                log.warning('Please install mmdc to generate mermaid images.')
            md_pipe = f"```mermaid\n{info['mermaid']['pipe']}```"
            md_perf = f"```mermaid\n{info['mermaid']['perf']}```"

            r_str = f'```json\n{json.dumps(data, indent=4, ensure_ascii=False)}\n```'
            md_content = f'## result\n{r_str}\n## Pipeline\n{md_pipe}\n## Perfermence\n{md_perf}'
            md_file = f'logs/{fname}_report.md'
            with open(md_file, 'w') as f: f.write(md_content)

        log.debug(f'pipe detail:\n{json.dumps(info, indent=4, ensure_ascii=False)}')
        info['logs'] = []
        for k in pipe_manager:
            info['logs'] += pipe_manager[k].inout_log
        
        return info

    async def async_run(self, data, save_perf=False):
        start_t = time.time()
        result = await self.pipetree.async_run(data)
        log.debug(f'final out:\n{json.dumps(result, indent=4, ensure_ascii=False)}')
        info = self.gen_info(result, start_t, save_perf)
        return result, info

    def seq_run(self, prompt_id, ws_id, task_data, send_msg):
        task_order = get_ordered_task(task_data)
        task_out = {}

        for task_id in task_order:
            msg = {
                "node": task_id, 
                "display_node": task_id,
                "prompt_id": prompt_id
            }
            send_msg("executing", msg, ws_id)
            task = task_data[task_id]
            task_type = task['class_type']

            if task_type == 'InputText':
                task_out[task_id] = [task['inputs']['input']]
            elif task_type == 'PreviewText':
                inp_id, inp_idx = task['inputs']['out'][0].split('-')
                inp_idx = int(inp_idx)
                out = task_out[inp_id][inp_idx] if inp_id in task_out else None

                msg = {
                    "node": task_id,
                    "display_node": task_id,
                    "output": {
                        "string": [out],
                    },
                    "prompt_id": prompt_id
                }
                send_msg("executed", msg)
            elif 'prompt' in task['inputs']:
                prompt = task['inputs']['prompt']
                for k, v in task['inputs'].items():
                    if k == 'prompt': continue
                    elif k == '模型' or k == 'model': continue
                    else:
                        inp_id, inp_idx = v[0].split('-')
                        inp_idx = int(inp_idx)
                        inp = task_out[inp_id][inp_idx] if inp_id in task_out else None
                        prompt = prompt.replace(k, inp)
                out = self.llm_backend(prompt)
                task_out[task_id] = [out]

                msg = {
                    "node": task_id,
                    "display_node": task_id,
                    "output": {
                        "string": [out],
                    },
                    "prompt_id": prompt_id
                }
                send_msg("executed", msg)
            elif 'kb' in task['inputs']:
                ...
            elif 'use_llm' in task['inputs']:
                ...
            else:
                time.sleep(3)

        return msg

    def fake_run(self, prompt_id, ws_id, task_data, send_msg):
        # 模拟
        for node_id in [2, 3, 8, 5, 7, 6]:
        # for node_id in [4, 10, 11, 5, 13]:
            data = {
                "node": str(node_id), 
                "display_node": str(node_id),
                "prompt_id": prompt_id
            }
            send_msg("executing", data, ws_id)
            time.sleep(3)

        # for i in range(20):
        #     data = {
        #         "value": i+1,
        #         "max": 20,
        #         "prompt_id": prompt_id,
        #         "node": "13"
        #     }
        #     self.send_msg("progress", data, ws_id)
        #     time.sleep(.5)

        # for node_id in [12, 9]:
        #     data = {
        #         "node": str(node_id), 
        #         "display_node": str(node_id),
        #         "prompt_id": prompt_id
        #     }
        #     self.send_msg("executing", data, ws_id)
        #     time.sleep(1)

        # out = {
        #     "node": "9",
        #     "display_node": "9",
        #     "output": {
        #         "images": [{
        #             "filename": "doubao.png",
        #             "subfolder": "",
        #             "type": "temp"
        #         }],
        #     },
        #     "prompt_id": prompt_id
        # }

        out = {
            "node": "6",
            "display_node": "6",
            "output": {
                "string": ["1234dededexdeee\ndefwefew"],
            },
            "prompt_id": prompt_id
        }
        send_msg("executed", out)
        return out

    def _run(self, data, save_perf=False, core_num=4):
        start_t = time.time()
        match self.run_mode:
            case 'async':
                log.debug(f"Run '{self.name}' pipeline in coroutine")
                result = asyncio.run(self.pipetree.async_run(data))
            case 'mp':
                log.debug(f"Run '{self.name}' pipeline in multiprocess")
                result = self.pipetree.mp_run(data, core_num)
            case _:
                log.debug(f"Run '{self.name}' pipeline in sequential")
                result = self.pipetree.normal_run(data)
        log.debug(f'final out:\n{json.dumps(result, indent=4, ensure_ascii=False)}')
        info = self.gen_info(result, start_t, save_perf)
        return result, info

    def run(self, data, save_perf=False, core_num=4, split=None):
        if (t := type(data)) is dict:
            return self._run(data, save_perf, core_num)
        elif t is list:
            if self.run_mode == 'async':
                async def f():
                    results = []
                    if split is None:
                        tasks = []
                        for d in data:
                            task = asyncio.create_task(self.pipetree.async_run(d))
                            tasks.append(task)
                        results = await tqdm_asyncio.gather(*tasks)
                    else:
                        parts = len(data) // split + 1
                        for i in tqdm.trange(parts):
                            tasks = []
                            for d in data[i*split:(i+1)*split]:
                                task = asyncio.create_task(self.pipetree.async_run(d))
                                tasks.append(task)
                            results += await asyncio.gather(*tasks)
                    return results

                results = asyncio.run(f())
                return [(r, None) for r in results]
            else:
                all_result = [self._run(d, save_perf, core_num) for d in tqdm.tqdm(data)]
            return all_result

    async def replay(self, node_name, data_arr):
        node = self.pipetree.node_manager[node_name]
        pipe = self.pipetree.pipe_manager.get(node_name, None)

        tasks = []
        for data in data_arr:
            task = asyncio.create_task(node.replay(data))
            tasks.append(task)
        await asyncio.gather(*tasks)

        ret = []
        for data in data_arr:
            d = {}
            for o in node.mermaid_outs:
                d[o] = data[o]
            ret.append(d)
        return ret, len(pipe.run_time) if pipe else node.run_cnt

    def to_png(self, pipe_img):
        if check_cmd_exist('mmdc'):
            pipe_mermaid = self.pipetree.tree2mermaid()
            tmp_file = f'/tmp/{uuid.uuid4()}'
            with open(tmp_file, 'w') as f: f.write(pipe_mermaid)
            log.debug(f'Save pipeline mermaid in: {tmp_file}')
            os.popen(f'mmdc -i {tmp_file} -o {pipe_img} -s 3 >/dev/null 2>&1')
            log.debug(f'save {pipe_img}')
        else:
            log.warning('Please install mmdc to generate mermaid images.')

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.name}, mode: {self.run_mode}, file: {self.pipefile}, hash: {self.hash[-8:]}>"

    def __repr__(self):
        return self.__str__()


from ollama import Client, AsyncClient
from . import *

ollama_model = os.getenv('OLLAMA_MODEL')
ollama_num_ctx = int(os.getenv('OLLAMA_NUM_CTX', 2048))

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    client = Client(timeout=6000)
    resp = client.chat(model=ollama_model, messages=[
        {
            'role': 'user',
            'content': text,
        },
    ], options={
        "num_ctx": ollama_num_ctx
    })

    return resp.message.content

async def async_completion(text):
    client = AsyncClient(timeout=6000)

    if (t := type(text)) is str:
        msg = [
            {
                "role": "user",
                "content": text,
            }
        ]
    elif t is list:
        msg = []
        for role, c in zip(['user', 'assistant']*len(text), text):
            msg.append({
                "role": role,
                "content": c,
            })

    async with llm_sem:
        resp = await client.chat(
            model=ollama_model,
            messages=msg,
            options={
                "num_ctx": ollama_num_ctx
            }
        )

    return resp.message.content



import re
import time
import json
from .log import log
from .clients import SearchEngine

class Pipe:
    def __init__(self, name, lock, run_time, inout_log, verbose, retry=1, second_round=False):
        self.name = name
        self.verbose = verbose
        self.retry = retry
        self.second_round = second_round
        self.log = lambda n, t: log.debug(f'[{name}] {n}: {t}') if self.verbose else None

        # multiprocess lock
        self.lock = lock
        self.run_time = run_time if run_time is not None else []
        self.inout_log = inout_log if inout_log is not None else []

    @property
    def time(self):
        if self.run_time:
            return sum(self.run_time) / len(self.run_time)
        return None

    def _json(self, inp):
        try:
            j = json.loads(inp)
            return j
        except: pass
        
        if (m := re.findall(r'```(?:json)?(.*?)```', inp, re.DOTALL)):
            try:
                j = json.loads(m[0])
                return j
            except: pass
        return None

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.name}>"

    def __repr__(self):
        return self.__str__()

    def _call(self, *inp):
        pass

    def __call__(self, *inp):
        n = 0
        while n < self.retry:
            start_t = time.time()
            self.log('inp', inp)
            out, query, resp = self._call(*inp)

            t = time.time() - start_t
            inout = {
                'id': self.name,
                'timestamp': time.time(),
                'input': query,
                'output': resp,
            }
            self.log('cost time', t)
            if self.lock is not None:
                with self.lock:
                    self.run_time.append(t)
                    self.inout_log.append(inout)
            else:
                self.run_time.append(t)
                self.inout_log.append(inout)
            if out is None:
                n += 1
                self.log('retry', n)
            else:
                break
        return out

    async def _async_call(self, *inp):
        pass

    async def _async_second_call(self, history):
        pass

    async def async_call(self, *inp):
        n = 0
        while n < self.retry:
            start_t = time.time()
            self.log('inp', inp)
            out, query, resp = await self._async_call(*inp)

            t = time.time() - start_t
            inout = {
                'id': self.name,
                'timestamp': time.time(),
                'input': query,
                'output': resp,
            }
            self.log('cost time', t)
            if self.lock is not None:
                with self.lock:
                    self.run_time.append(t)
                    self.inout_log.append(inout)
            else:
                self.run_time.append(t)
                self.inout_log.append(inout)
            
            if out is None and self.second_round:
                start_t = time.time()
                out, query2, resp2 = await self._async_second_call([query, resp])

                t = time.time() - start_t
                inout = {
                    'id': self.name,
                    'timestamp': time.time(),
                    'input': query2,
                    'output': resp2,
                }
                self.log('cost time', t)
                if self.lock is not None:
                    with self.lock:
                        self.run_time.append(t)
                        self.inout_log.append(inout)
                else:
                    self.run_time.append(t)
                    self.inout_log.append(inout)
            
            if out is None:
                n += 1
                self.log('retry', n)
            else:
                break
        return out

class LLMPipe(Pipe):
    def __init__(self, name, prompt=None, return_json=True, format=None, llm=None, verbose=True, retry=5, inp=None, out=None, lock=None, run_time=None, inout_log=None, second_round=False, **kargs):
        if return_json: second_round = True
        super().__init__(name, lock, run_time, inout_log, verbose, retry, second_round)
        self.prompt = prompt
        self.llm = llm
        self.return_json = return_json
        self.format = format

    def _call(self, *inp):
        text = self.prompt(*inp)
        self.log('prompt', text)
        resp = self.llm(text)
        self.log('resp', resp)
        if self.return_json:
            out = self._json(resp)
            self.log('json', out)

            if out and type(out) is dict and self.format is not None:
                if type(self.format) is dict and all(map(lambda x: x[0] in out and (type(out[x[0]]) is x[1] if type(x[1]) is type else out[x[0]] in x[1]), self.format.items())):
                    self.log(f'check {self.format}', '✓')
                elif type(self.format) is set and all(i in out for i in self.format):
                    self.log(f'check {self.format}', '✓')
                else:
                    self.log(f'check {self.format}', '✗')
                    out = None
        else:
            out = resp
        return out, text, resp

    async def _async_call(self, *inp):
        text = self.prompt(*inp)
        self.log('prompt', text)
        resp = await self.llm(text)
        self.log('resp', resp)
        if self.return_json:
            out = self._json(resp)
            self.log('json', out)

            if out and type(out) is dict and self.format is not None:
                if type(self.format) is dict and all(map(lambda x: x[0] in out and (type(out[x[0]]) is x[1] if type(x[1]) is type else out[x[0]] in x[1]), self.format.items())):
                    self.log(f'check {self.format}', '✓')
                elif type(self.format) is set and all(i in out for i in self.format):
                    self.log(f'check {self.format}', '✓')
                else:
                    self.log(f'check {self.format}', '✗')
                    out = None
        else:
            out = resp
        return out, text, resp
    
    async def _async_second_call(self, history):
        if self.return_json:
            history.append('Make sure return answer in JSON format.')

            self.log('history', history)
            resp = await self.llm(history)
            self.log('resp', resp)

            out = self._json(resp)
            self.log('json', out)

            if out and type(out) is dict and self.format is not None:
                if type(self.format) is dict and all(map(lambda x: x[0] in out and type(out[x[0]]) is x[1], self.format.items())):
                    self.log(f'check {self.format}', '✓')
                elif type(self.format) is set and all(i in out for i in self.format):
                    self.log(f'check {self.format}', '✓')
                else:
                    self.log(f'check {self.format}', '✗')
                    out = None
            return out, history, resp

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.name}, prompt: {self.prompt.name}, json: {self.return_json}>"

class RAGPipe(Pipe):
    def __init__(self, name, rag=None, verbose=True, return_key=None, lock=None, run_time=None, inout_log=None, **kargs):
        super().__init__(name, lock, run_time, inout_log, verbose)
        self.rag = rag
        self.return_key = return_key

    def _call(self, *inp):
        resp = self.rag(*inp)
        self.log('out', resp)
        if resp is not None and len(resp):
            if self.return_key:
                o = eval(f'resp{self.return_key}')
                self.log('return_key', o)
                out = o
            else:
                out = resp
        return out, inp, resp

    async def _async_call(self, *inp):
        resp = await self.rag(*inp)
        self.log('out', resp)
        if resp is not None and len(resp):
            if self.return_key:
                o = eval(f'resp{self.return_key}')
                self.log('return_key', o)
                out = o
            else:
                out = resp
        return out, inp, resp

class SearchPipe(Pipe):
    def __init__(self, name, search_engine='bing', count=5, verbose=True, lock=None, run_time=None, inout_log=None, **kargs):
        super().__init__(name, lock, run_time, inout_log, verbose)
        self.count = count
        if type(search_engine) is str:
            self.engine, self.async_engine = SearchEngine.get(search_engine)
        else:
            self.engine = self.async_engine = search_engine

    def _call(self, inp):
        resp = self.engine(inp, count=self.count)
        self.log('out', resp)
        return resp, inp, resp

    async def _async_call(self, inp):
        resp = await self.async_engine(inp, count=self.count)
        self.log('out', resp)
        return resp, inp, resp

class BrowserPipe(Pipe):
    def __init__(self, name, client=None, verbose=True, lock=None, run_time=None, inout_log=None, **kargs):
        super().__init__(name, lock, run_time, inout_log, verbose)
        self.client = client

    def _call(self, inp):
        resp = self.rag(inp)
        self.log('out', resp)
        return resp, inp, resp

    async def _async_call(self, inp):
        resp = await self.rag(inp)
        self.log('out', resp)
        return resp, inp, resp



import json
from .log import log

class PromptKeys:
    def __init__(self, arr, prompt):
        self._arr = arr.copy()
        self.prompt = prompt

    def __getitem__(self, index):
        return self._arr[index]

    def __setitem__(self, index, value):
        log.debug(f'Change prompt "{self.prompt.name}" keys [{index}] to: {value}')
        self._arr[index] = value
        self.prompt.save()

    def __str__(self):
        return str(self._arr)

class Prompt:
    def __init__(self, text, keys, name, file=None):
        self.name = name
        self.file = file
        self.__text = text
        self.__keys = PromptKeys(keys, self)

    @property
    def text(self):
        return self.__text

    @text.setter
    def text(self, value):
        log.debug(f'Change prompt "{self.name}" text to: {value}')
        self.__text = value
        self.save()

    @property
    def keys(self):
        return self.__keys[:]

    @keys.setter
    def keys(self, value):
        log.debug(f'Change prompt "{self.name}" keys to: {value}')
        self.__keys = value
        self.save()

    def __call__(self, *inps):
        t = self.__text
        for a, b in zip(self.__keys, inps):
            if b is None: b = str(b)
            elif type(b) is not str: b = json.dumps(b, ensure_ascii=False)
            t = t.replace(a, b)
        return t

    def save(self):
        if self.file is not None:
            log.debug(f'Save prompt [{self.name}] to: {self.file}')
            with open(self.file, 'w') as f:
                f.write(f'prompt="""{self.text}"""\nkeys={self.keys}')



import os
import logging
import datetime
from pathlib import Path
from rich.console import Console
from rich.logging import RichHandler
from logging.handlers import RotatingFileHandler

def beijing(sec, what):
    beijing_time = datetime.datetime.utcnow() + datetime.timedelta(hours=8)
    return beijing_time.timetuple()

LOG_DIR = os.getenv('LOGGING_DIR', 'logs')
LOG_LEVEL = os.getenv('LOGGING_LEVEL', 'NOTSET')
MAXBYTES = 10000000 # ~10M
LOGFORMAT = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
LOGFORMAT_RICH = '%(message)s'
rh = RichHandler(console=Console(stderr=True))
rh.setFormatter(logging.Formatter(LOGFORMAT_RICH))
log_dir = Path(LOG_DIR)
log_dir.mkdir(exist_ok=True, parents=True)
log_file = log_dir / f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}.log"
logging.Formatter.converter = beijing
logging.basicConfig(level=LOG_LEVEL, format=LOGFORMAT, datefmt="[%y-%m-%d %H:%M:%S]", handlers=[rh, RotatingFileHandler(log_file, maxBytes=MAXBYTES)])
log = logging.getLogger("llmpipeline")

def _banner_print(text):
    n = len(text)
    t = '\n' + '-'*(4+n) + '\n| ' + text + ' |\n' + '-'*(4+n)
    log.debug(t)

log.banner = _banner_print


import os
import asyncio
import pkgutil
import importlib
from ..log import log

class _SearchEngine:
    def __init__(self):
        log.debug('Setup SearchEngine')
        self.engines = {}
        for loader, module_name, is_pkg in pkgutil.walk_packages(__path__, __name__ + '.'):
            if (name := module_name.split('.')[-1]).startswith('search_'):
                name = name.replace('search_', '')
                m = importlib.import_module(module_name)
                self.engines[name] = [m.search, m.async_search]
        log.debug(f'Load engines: {list(self.engines.keys())}')

    def get(self, engine_name):
        return self.engines[engine_name]

SearchEngine = _SearchEngine()
llm_sem = asyncio.Semaphore(int(os.getenv('LLM_CONCURRENCY', 16)))


import sys
from pathlib import Path
from .log import log
from .utils import *
from .prompt import Prompt
from .pipeline import Pipeline

class PromptManager:
    def __init__(self, prompts_dir=None):
        log.debug('Setup PromptManager')
        self.buildin_prompts_dir = Path(__file__).parent / 'build_in_prompts'
        if prompts_dir is None:
            log.debug('PromptManager dir is None, set to current dir.')
            self.prompts_dir = Path('.')
        else:
            self.prompts_dir = Path(prompts_dir)
        self.prompts = {}
        self.load_prompts()

    def load_prompts(self):
        self.prompts = {}
        prompts_dirs = [self.buildin_prompts_dir, self.prompts_dir]
        log.debug(f'Start load prompts: {prompts_dirs}')

        for pdir in prompts_dirs:
            prompt_files = list(pdir.glob('*_prompt.py'))
            log.debug(f'Find {len(prompt_files)} prompt files: {[p.stem for p in prompt_files]}')

            for pf in prompt_files:
                m = importpath(pf)
                self.prompts[pf.stem] = Prompt(m.prompt, m.keys, pf.stem, pf)

        log.debug('All prompts loaded')
    
    def get(self, item):
        if (t := type(item)) is str:
            if item not in self.prompts: raise KeyError(f"Don't has prompt: {item}")
            return self.prompts[item]
        elif t is dict:
            name = f'prompt_{len(self.prompts)}'
            assert name not in self.prompts, f'PromptManager: two prompts has the same name {name}'
            self.prompts[name] = Prompt(item['prompt'], item['keys'], name)
            return self.prompts[name]

class PipelineManager:
    def __init__(self, llm_client=None, rag_client=None, run_mode='async', pipes_dir=None, prompt_manager=None):
        log.debug('Setup PipelineManager')
        if pipes_dir is None: pipes_dir = '.'
        self.pipes_dir = Path(pipes_dir)
        self.prompt_manager = prompt_manager or PromptManager()
        self.llm_client = llm_client
        self.rag_client = rag_client
        self.run_mode = run_mode
        self.pipes = {}
        self.load_pipes()

    def load_pipes(self):
        self.pipes = {}
        log.debug(f'Start load pipelines: {self.pipes_dir}')
        sys.path.append(str(self.pipes_dir))
        pipe_files = list(self.pipes_dir.glob('*_pipeline.py'))
        log.debug(f'Find {len(pipe_files)} pipeline files: {[p.stem for p in pipe_files]}')

        for pf in pipe_files: self._load_pipe(pf.stem, pipefile=pf)
        sys.path.pop()

        log.debug('All pipelines loaded')

    def add_pipe(self, name, pipeconf=None, pipefile=None, run_mode=None):
        if pipefile:
            if type(pipefile) is str: pipefile = Path(pipefile)
            sys.path.append(str(pipefile.parent))

        p = self._load_pipe(name, pipeconf=pipeconf, pipefile=pipefile, run_mode=run_mode)

        if pipefile: sys.path.pop()

        return p

    def _load_pipe(self, name, pipeconf=None, pipefile=None, run_mode=None):
        if pipefile and name in self.pipes and calc_sha256(pipefile) == self.pipes[name].hash:
            p = self.pipes[name]
        else:
            p = Pipeline(self.llm_client, self.rag_client, self.prompt_manager, pipeconf=pipeconf, pipefile=pipefile, run_mode=run_mode or self.run_mode)
            self.pipes[name] = p
        return p

    def export_pipe_conf(self):
        conf = {k: copy.deepcopy(v['conf']) for k,v in self.pipes.items()}
        for c in conf.values():
            for pipe_conf in c.values():
                if 'format' in pipe_conf:
                    for k, v in pipe_conf['format'].items():
                        pipe_conf['format'][k] = str(v)
        return conf

    def update_pipe(self, pipe_name, pipe_data):
        if pipe_name.endswith('_pipe'):
            pipe_file = self.pipes_dir / f'{pipe_name}.py'

            try:
                content = json.dumps(pipe_data, indent=4, ensure_ascii=False)
                content = re.sub(r'"<class \'(.*)\'>"', r'\1', content)

                conf = pipe_data
                for pipe_conf in conf.values():
                    if 'format' in pipe_conf:
                        for k, v in pipe_conf['format'].items():
                            pipe_conf['format'][k] = eval(v.replace("<class '","").replace("'>",""))

                pipe = copy.deepcopy(conf)
                for k in pipe:
                    if 'prompt' in pipe[k]:
                        pipe[k]['prompt'] = self.prompt_manager.prompts[pipe[k]['prompt']]

                self.pipes[pipe_name] = {
                    'file': pipe_file,
                    'conf': conf,
                    'func': Pipeline(pipe, self.llm_client, self.rag_client)
                }

                with open(pipe_file, 'w') as f:
                    content = 'pipe = ' + content + '\n'
                    f.write(content)
                
                log.debug(f'Save pipeline "{pipe_name}":\n{content}')
                return f'Successfully saved {pipe_name}'
            except Exception as e:
                return f'Error: {e}'
        else:
            return 'Error: name must end with `_pipe`.'

    def export_nodes(self):
        nodes = {}
        for pipeline in self.pipes.values():
            pt = pipeline.pipetree
            for node in pt.node_manager.values():
                nodes |= node.export_as_comfyui()
        return nodes

    def __str__(self):
        return f"<{self.__class__.__name__} mode: {self.run_mode}, pipes: {list(self.pipes)}, dir: {self.pipes_dir.absolute()}>"

    def __repr__(self):
        return self.__str__()



import os
from fastapi import APIRouter
from pydantic import BaseModel
from fastapi import HTTPException
from . import PromptManager, PipelineManager

class PromptData(BaseModel):
    name:   str | None = None
    text:   str | None = None
    keys:  list | None = None

class PipeData(BaseModel):
    name:   str | None = None
    data:  dict | None = None

class PipelineAPI:
    def __init__(self, prompt_dir, pipeline_dir, llm_client=None, rag_client=None, is_async=False):
        router = APIRouter(prefix='/api')
        prompt_manager = PromptManager(prompt_dir)
        pipeline_manager = PipelineManager(llm_client, rag_client, is_async, pipeline_dir, prompt_manager)

        @router.get("/prompt/list")
        async def prompt_list():
            try:
                ret = {k: {'text': v.text, 'keys': v.keys} for k,v in prompt_manager.prompts.items()}
                return ret

            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        @router.post("/prompt/update")
        async def prompt_update(p_data: PromptData):
            try:
                if p_data.text:
                    prompt_manager.prompts[p_data.name].text = p_data.text
                if p_data.keys:
                    prompt_manager.prompts[p_data.name].keys = p_data.keys

                ret = {
                    p_data.name: {
                        'text': prompt_manager.prompts[p_data.name].text,
                        'keys': prompt_manager.prompts[p_data.name].keys
                    }
                }
                return ret

            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))


        @router.get("/pipeline/list")
        async def pipeline_list():
            try:
                ret = pipeline_manager.export_pipe_conf()
                return ret

            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        @router.post("/pipeline/mermaid")
        async def pipeline_mermaid(p_data: PipeData):
            try:
                pipeline = pipeline_manager.pipes[p_data.name]['func']
                mermaid = pipeline.pipe2mermaid(pipeline.pipe)
                ret = {
                    'mermaid': mermaid,
                }
                return ret

            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        @router.post("/pipeline/update")
        async def pipeline_update(p_data: PipeData):
            try:
                ret = {
                    'result': pipeline_manager.update_pipe(p_data.name, p_data.data)
                }
                return ret

            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        self.router = router
        self.prompt_manager = prompt_manager
        self.pipeline_manager = pipeline_manager



from . import *
import json
import httpx
from fastapi import HTTPException

json_file = os.getenv('RAG_JSON')
data = json.load(open(json_file, 'r'))

def rag_client(is_async=False):
    if is_async:
        return async_rag
    else:
        return rag

def rag(text):
    return data.get(text, None)

async def async_rag(text):
    return rag(text)



from . import *
import json
import httpx
import requests
from fastapi import HTTPException

import warnings
from urllib3.exceptions import InsecureRequestWarning
warnings.simplefilter("ignore", InsecureRequestWarning)

def rag_client(is_async=False):
    if is_async:
        return async_rag
    else:
        return rag

def rag(text, url=None, index=None):
    j = {
        'type': index,
        'query': json.dumps([str(text)])
    }

    response = requests.post(url, json=j, verify=False)

    if response.status_code != 200:
        raise Exception(f"Failed to generate completion: {response.text}")
    
    pids = [i['condiate_result'] for i in response.json()[0]['condidate']]
    return pids

async def async_rag(text):
    messages = [
        {
            'action': 'From user', # 'To user'
            'content': text,
        }
    ]

    async with llm_sem:
        input_data = {
            "action": "To user",
            "parent_messages": messages,
            "gen_kwargs": {
                "model": pulse_model,
                "num_return_sequences": 1,
                "temperature": temperature,
                "top_p": top_p,
                "top_k": top_k,
                "max_tokens": max_tokens,
                "repetition_penalty": repetition_penalty,
            }
        }

        return await _get_completion(
            input_data=input_data,
            retry=retry_count
        )

async def _get_completion(input_data: dict, retry: int = 5):
    try:
        async with httpx.AsyncClient(verify=False) as client:
            async with client.stream(
                "POST",
                pulse_url,
                json=input_data,
                headers={
                    "Authorization": f"Bearer {pulse_token}",
                },
                timeout=600,
            ) as stream:
                # error
                if stream.status_code != 200:
                    error_detail = ""
                    try:
                        error_detail = (await stream.aread()).decode("utf8")
                        error_detail = str(json.loads(error_detail))
                    except:
                        pass

                    raise HTTPException(status_code=500, detail=error_detail)

                content = (await stream.aread()).decode("utf8")
                content = json.loads(content)
                content = "".join(content['messages'][0]['content']['parts'])
                return content
    except Exception as e:
        # logger.error(json.dumps({
        #     "retry": retry,
        #     "input": input_data,
        #     "error": traceback.format_exc()
        # }, ensure_ascii=False, indent=4))
        # 重试次数到达极限
        if retry == 0: raise e

        return await _get_completion(
            input_data=input_data,
            retry=retry-1,
        )



import re
import copy
import time
import queue
import asyncio
import importlib
import traceback
import collections
from enum import Enum
from pathlib import Path
from functools import reduce
import multiprocessing as mp
from .log import log
from .prompt import Prompt
from .pipe import *
import pymupdf4llm

class DataState(Enum):
    VOID = 0

class Color:
    Black = 'black'
    White = 'white'
    Gray = '#ECE4E2'
    Pink = '#FE929F'
    RED = '#D64747'
    LightPink = '#FAB6BF'
    Khaki = '#CC8A4D'
    DarkBlue = '#445760'
    LightGreen = '#EAFFD0'
    Green = '#9BCFB8'
    LightYellow = '#FFFFAD'
    Black2 = '#3D3E3F'

class NodeColorStyle:
    default = f'color:{Color.Black}'
    LLMNode = f'fill:{Color.Gray},color:{Color.Black}'
    RAGNode = f'fill:{Color.Pink},color:{Color.Black}'
    LoopNode = f'fill:none,stroke:{Color.Khaki},stroke-dasharray:5 5,stroke-width:2px'
    BranchNode = f'fill:{Color.DarkBlue},color:{Color.White}'
    CodeNode = f'fill:{Color.LightYellow},color:{Color.Black}'
    WebNode = f'fill:{Color.LightPink},color:{Color.Black}'
    ValueNode = f'fill:{Color.LightGreen},color:{Color.Black}'
    ExitNode = f'fill:{Color.Black2},color:{Color.White}'
    FileNode = f'fill:{Color.Khaki},color:{Color.Black}'
    Data = f'fill:{Color.Green},color:{Color.Black}'
    InputData = f'fill:{Color.RED},color:{Color.Black}'

class NodeShape:
    default = lambda x: f'{x}["{x}"]' # 矩形
    LLMNode = lambda x: f'{x}["{x}"]'
    RAGNode = lambda x: f'{x}("{x}")' # 圆角矩形
    LoopNode = lambda x: f'{x}(("{x}"))' # 圆形
    BranchNode = lambda x: f'{x}{{"{x}"}}'
    CodeNode = lambda x: f'{x}[/"{x}"/]'
    WebNode = lambda x: f'{x}("{x}")'
    ValueNode = lambda n, x: f'{n}{{{{"{x}"}}}}'
    ExitNode = lambda x: f'{x}[["{x}"]]'
    FileNode = lambda x: f'{x}["{x}"]'
    Data = lambda x: f'{x}(["{x}"])'
    InputData = lambda x: f'{x}(["{x}"])'

class Data:
    mermaid_style = NodeColorStyle.Data
    mermaid_shape = NodeShape.Data

class InputData:
    mermaid_style = NodeColorStyle.InputData
    mermaid_shape = NodeShape.InputData

class Node:
    mermaid_style = NodeColorStyle.default
    mermaid_shape = NodeShape.default
    mermaid_inline = '-->'
    mermaid_outline = '-.->'
    mermaid_toexit = '--o'
    mermaid_inline_passed = '==>'
    mermaid_outline_passed = '==>'
    mermaid_toexit_passed = '==o'

    def __init__(self, name, conf, tree=None):
        self.name = name
        self.conf = conf
        self.tree = tree
        self.run_cnt = 0
        self.max_cnt = self.conf.get('max_cnt', None)
        self.in_loop = None
        self.next = []
        self.loop_nodes = []
        self.reset_out_flag = 'reset_out' in self.conf
        if self.reset_out_flag:
            self.conf['out'] = self.conf['reset_out']
            del self.conf['reset_out']
        self.set_mermaid()
        self.post_init()

    def set_mermaid(self):
        self.mermaid_inps = []
        if 'inp' in self.conf:
            for i in self.conf['inp']:
                if (t := type(i)) is str:
                    self.mermaid_inps.append(i)
                elif t is dict:
                    self.mermaid_inps += list(i.values())

        outs = []
        if (o := self.conf.get('out', None)):
            if (t := type(o)) is str:
                outs = [o]
            elif t is list:
                outs = o[:]
            elif t is dict:
                outs = list(o.values())
        self.mermaid_outs = outs
        self.mermaid_data = self.mermaid_inps + self.mermaid_outs

    def _get_mermaid_defines(self):
        return [self.__class__.mermaid_shape(self.name)] + [Data.mermaid_shape(d) for d in self.mermaid_data]

    def get_mermaid(self, info=None):
        inps = ' & '.join(self.mermaid_inps) or None
        outs = ' & '.join(self.mermaid_outs) or None

        defines = self._get_mermaid_defines()

        t = info["detail"][self.name]["avg_time"] if info and self.name in info['detail'] else None
        if t is not None: t = f'|{t:.2f}s|'

        inline = self.mermaid_inline_passed if self.run_cnt else self.mermaid_inline
        outline = self.mermaid_outline_passed if self.run_cnt else self.mermaid_outline
        inout_link = (inps, inline, self.name, outline, t, outs, None)

        links = [inout_link]
        for n in self.next:
            if n.name == 'exit':
                t = f'|total: {info["total_time"]:.2f}s|' if info and self.name in info['exec_path'] else None
                links.append((None, None, outs, self.mermaid_toexit_passed if t else self.mermaid_toexit, t, 'exit', None))

        subg = []

        return defines, links, subg

    def update(self, nodes):
        for name in self.conf.get('next', []):
            if name in nodes: self.next.append(nodes[name])

    def post_init(self):
        pass

    def reset(self):
        self.run_cnt = 0

    def get_inps_mp(self, data, config=None):
        def get_data(i):
            if i not in data: return DataState.VOID
            if type(data[i]) is list:
                if config and i in config['loop_index']:
                    loop_i = config['loop_index'][i]
                    return data[i][loop_i]
                elif DataState.VOID in data[i]:
                    return DataState.VOID
            return data[i]

        inps = []
        for i in self.conf['inp']:
            if type(i) is str:
                if (d := get_data(i)) is DataState.VOID: return []
                else: inps.append(d)
            elif type(i) is dict:
                t = {}
                for k, v in i.items():
                    if (d := get_data(v)) is DataState.VOID: return []
                    else: t[k] = d
                inps.append(t)
        return inps

    def current_mp_task(self, inps, data, queue, config=None):
        if self.__class__.__name__ in ['LLMNode', 'RAGNode']:
            out = self.pipe(*inps)
            self.set_out(out, data, config=config)

        for n in self.next: queue.put((n.name, config))

    def mp_run(self, mp_name, data, queue, perf, config=None):
        if (inps := self.get_inps_mp(data, config)):
            start_time = time.time()
            self.run_cnt += 1
            cnt = self.run_cnt
            log.banner(f"Enter mp task: {self.name}, cnt: {cnt}, {mp_name}")
            self.current_mp_task(inps, data, queue, config)
            log.banner(f"Leave mp task: {self.name}, cnt: {cnt}, {mp_name}")
            perf.put((mp_name, self.name, start_time, time.time()))
        else:
            queue.put((self.name, config))

    async def get_inps(self, queue):
        inps = []
        for i in self.conf['inp']:
            if type(i) is str:
                d = await queue[i].get()
                inps.append(d)
                queue[i].put_nowait(d)
            elif type(i) is dict:
                t = {}
                for k, v in i.items():
                    d = await queue[v].get()
                    t[k] = d
                    queue[v].put_nowait(d)
                inps.append(t)
        return inps

    def set_out(self, out, data, queue=None, config=None):
        def set_data(k, v, config, queue):
            if config:
                with self.tree.mp_lock:
                    i = config['loop_index'][k]
                    pre = data[k]
                    pre[i] = v
                    data[k] = pre
            else:
                data[k] = v

            if queue:
                arr = [queue]
                while arr:
                    q = arr.pop(0)
                    q[k].put_nowait(v)
                    if '_sub' in q: arr += q['_sub']

        o = self.conf['out']
        if (t := type(o)) is str:
            set_data(o, out, config, queue)
        elif t is list:
            for k in o:
                set_data(k, out[k], config, queue)
        elif t is dict:
            if out is not None and type(out) is dict:
                for k in o:
                    set_data(o[k], out.get(k, None), config, queue)
            else:
                for k in o: set_data(o[k], None, config, queue)

    def reset_out(self, queue):
        def q_del(q, k):
            if k in q:
                while not q[k].empty(): q[k].get_nowait()
            if '_sub' in q:
                for sub_q in q['_sub']: q_del(sub_q, k)

        o = self.conf['out']
        if (t := type(o)) is str:
            q_del(queue, o)
        elif t is list:
            for k in o: q_del(queue, k)
        elif t is dict:
            for k in o: q_del(queue, o[k])

    async def add_task(self, data, queue, dynamic_tasks):
        for n in self.next:
            task = asyncio.create_task(n.run(data, queue, dynamic_tasks), name=n.name)
            dynamic_tasks.append(task)

    async def current_task(self, data, queue, dynamic_tasks):
        if self.__class__.__name__ in ['LLMNode', 'RAGNode']:
            inps = await self.get_inps(queue)
            out = await self.pipe.async_call(*inps)
            self.set_out(out, data, queue)

    async def async_run(self, data, queue, dynamic_tasks):
        start_time = time.time()
        self.run_cnt += 1
        if self.max_cnt is not None and self.run_cnt > self.max_cnt:
            log.banner(f"Async task: {self.name} hit max_cnt ({self.max_cnt}) limit, exit!")
        else:
            cnt = self.run_cnt
            log.banner(f"Enter async task: {self.name}, cnt: {cnt}")
            if self.reset_out_flag: self.reset_out(queue)
            await self.add_task(data, queue, dynamic_tasks)
            await self.current_task(data, queue, dynamic_tasks)
            log.banner(f"Leave async task: {self.name}, cnt: {cnt}")
            self.tree.perf.append(('coroutine', self.name, start_time, time.time()))

    async def replay(self, data):
        self.run_cnt += 1
        cnt = self.run_cnt
        log.banner(f"Enter async task: {self.name}, cnt: {cnt}")
        queue = collections.defaultdict(asyncio.Queue)
        for k, v in data.items(): queue[k].put_nowait(v)
        await self.current_task(data, queue, [])
        log.banner(f"Leave async task: {self.name}, cnt: {cnt}")

    def get_inps_normal(self, data):
        if 'inp' not in self.conf: return data
        inps = []
        for i in self.conf['inp']:
            if type(i) is str:
                if i not in data: return []
                else: inps.append(data[i])
            elif type(i) is dict:
                t = {}
                for k, v in i.items():
                    if v not in data: return []
                    else: t[k] = data[v]
                inps.append(t)
        return inps

    def current_normal_task(self, inps, data, queue):
        if self.__class__.__name__ in ['LLMNode', 'RAGNode']:
            out = self.pipe(*inps)
            self.set_out(out, data)

        for n in self.next: queue.append(n)

    def normal_run(self, data, queue):
        if (inps := self.get_inps_normal(data)):
            start_time = time.time()
            self.run_cnt += 1
            log.banner(f"Enter normal task: {self.name}, cnt: {self.run_cnt}")
            self.current_normal_task(inps, data, queue)
            log.banner(f"Leave normal task: {self.name}, cnt: {self.run_cnt}")
            self.tree.perf.append(('normal', self.name, start_time, time.time()))
        else:
            queue.append(self)

    @property
    def run(self):
        match self.tree.run_mode:
            case 'async':
                return self.async_run
            case 'mp':
                return self.mp_run
            case _:
                return self.normal_run

    def export_as_comfyui(self):
        return {}

    def __str__(self):
        return f"<{self.__class__.__name__}: {self.name}, next: {[n.name for n in self.next]}>"

    def __repr__(self):
        return self.__str__()

class LLMNode(Node):
    mermaid_style = NodeColorStyle.LLMNode
    mermaid_shape = NodeShape.LLMNode

    def post_init(self):
        tree = self.tree
        if type(self.conf['prompt']) is not Prompt:
            self.conf['prompt'] = tree.prompt_manager.get(self.conf['prompt'])
        
        if self.conf.get('backend', None):
            backend = self.conf['backend']
        elif (constructor := self.conf.get('backend_construct', None)):
            backend = constructor(tree.run_mode)
        else:
            backend = tree.llm_backend

        if tree.run_mode == 'mp':
            pipe = LLMPipe(
                    self.name,
                    llm=backend,
                    lock=tree.mp_lock,
                    run_time=tree.mp_manager.list(),
                    inout_log=tree.mp_manager.list(),
                    **self.conf
                    )
        else:
            pipe = LLMPipe(self.name, llm=backend, **self.conf)
        tree.pipe_manager[self.name] = pipe
        self.pipe = pipe

    def export_as_comfyui(self):
        inps = {i:["TEXT"] for i in self.pipe.prompt.keys}
        opt_inps = {"模型": ["MODEL"]}
        prompt = {
            "prompt": [
                "STRING",
                {"default": self.pipe.prompt.text, "multiline": True, "dynamicPrompts": True}
            ]
        }
        outs = self.mermaid_outs
        d = {
            "input": {
                "required": inps | prompt,
                "optional": opt_inps
            },
            "input_order": {"required": self.pipe.prompt.keys},
            "output": ["TEXT"] * len(outs),
            "output_is_list": [False] * len(outs),
            "output_name": outs,
            "name": self.name,
            "display_name": self.name,
            "description": f"{self.name} prompt",
            "python_module": "nodes",
            "category": "提示词",
            "output_node": False,
        }
        return {self.name: d}

class RAGNode(Node):
    mermaid_style = NodeColorStyle.RAGNode
    mermaid_shape = NodeShape.RAGNode

    def post_init(self):
        tree = self.tree
        if self.conf.get('backend', None):
            backend = self.conf['backend']
        elif (constructor := self.conf.get('backend_construct', None)):
            backend = constructor(tree.run_mode)
        else:
            backend = tree.rag_backend

        if (param := self.conf.get('rag_param', None)):
            rag_backend = lambda *x: backend(*x, **dict(param))
        else:
            rag_backend = backend

        if tree.run_mode == 'mp':
            pipe = RAGPipe(
                    self.name,
                    rag=rag_backend,
                    lock=tree.mp_lock,
                    run_time=tree.mp_manager.list(),
                    inout_log=tree.mp_manager.list(),
                    **self.conf
                    )
        else:
            pipe = RAGPipe(self.name, rag=rag_backend, **self.conf)

        tree.pipe_manager[self.name] = pipe
        self.pipe = pipe

    def export_as_comfyui(self):
        param = self.conf.get('rag_param', {})
        inps = {
            "text": ["TEXT"],
            "kb": ["STRING", {
                "default": param.get('kb_id', None),
                "multiline": False,
                "dynamicPrompts": True
            }],
            "top_k": ["INT", {
                "default": param.get('top_k', 1),
                "min": 1
            }],
            "threshold": ["FLOAT", {
                "default": param.get('threshold', 0.5),
                "min": 0.01,
                "max": 1.0,
                "step": 0.1
            }]
        }
        opt_inps = {}
        outs = self.mermaid_outs
        d = {
            "input": {
                "required": inps,
                "optional": opt_inps
            },
            "input_order": {"required": list(inps.keys())},
            "output": ["TEXT"] * len(outs),
            "output_is_list": [False] * len(outs),
            "output_name": outs,
            "name": self.name,
            "display_name": self.name,
            "description": f"{self.name} rag search",
            "python_module": "nodes",
            "category": "知识库",
            "output_node": False,
        }
        return {self.name: d}

class LoopNode(Node):
    mermaid_style = NodeColorStyle.LoopNode
    mermaid_shape = NodeShape.LoopNode

    def update(self, nodes):
        super().update(nodes)

        self.loop_nodes = []
        for name in self.conf['pipe_in_loop']:
            if name in nodes: self.loop_nodes.append(nodes[name])

        # inp = self.conf['inp'][0]
        # for n in self.loop_nodes:
        #     if inp in n.mermaid_inps:
        #         n.mermaid_inps[n.mermaid_inps.index(inp)] = self.name

    def get_mermaid(self, info=None):
        links = []
        inp = self.conf['inp'][0]
        links.append((None, None, inp, self.mermaid_inline, None, self.name, NodeColorStyle.LoopNode))

        defines = []
        subg = [(self.name, *self.conf['pipe_in_loop'])]

        return defines, links, subg

    def get_loop_outs(self):
        outs = []
        for n in self.loop_nodes:
            outs += n.mermaid_outs
        return outs

    async def current_task(self, data, queue, dynamic_tasks):
        inps = await self.get_inps(queue)
        inp = inps[0]
        n = len(inp)
        inp_name = self.conf['inp'][0]

        loop_tasks = []
        loop_data = []
        sub = []
        loop_outs = self.get_loop_outs()
        for item in inp:
            new_data = {}
            loop_data.append(new_data)

            new_queue = collections.defaultdict(asyncio.Queue)
            for k in queue:
                if k == inp_name:
                    new_queue[k].put_nowait(item)
                elif k not in loop_outs:
                    new_queue[k] = queue[k]
            sub.append(new_queue)

            for n in self.loop_nodes:
                task = asyncio.create_task(n.run(new_data, new_queue, loop_tasks))
                loop_tasks.append(task)

        queue['_sub'] = sub

        while not all(t.done() for t in loop_tasks):
            await asyncio.gather(*loop_tasks)

        del queue['_sub']
        for d in loop_data:
            for k, v in d.items():
                if k in data: data[k].append(v)
                else: data[k] = [v]
        for k in loop_data[0]: queue[k].put_nowait(data[k])

    def current_mp_task(self, inps, data, queue, config=None):
        N = len(inps[0])
        loop_outs = self.get_loop_outs()
        for k in loop_outs: data[k] = [DataState.VOID] * N
        for i in range(N):
            new_config = {} if config is None else copy.deepcopy(config)
            if 'loop_index' not in new_config: new_config['loop_index'] = {}
            new_config['loop_index'] |= {k:i for k in self.conf['inp']+loop_outs}
            for node in self.loop_nodes: queue.put((node.name, new_config))

        for node in self.next: queue.put((node.name, config))

    def current_normal_task(self, inps, data, queue):
        keys = list(data.keys())
        for item in inps[0]:
            tmp_d = copy.deepcopy(data)
            tmp_d[self.conf['inp'][0]] = item
            tmp_q = self.loop_nodes[:]
            while tmp_q:
                n = tmp_q.pop(0)
                n.run(tmp_d, tmp_q)

            for k in tmp_d:
                if k not in keys:
                    if k in data:
                        data[k].append(tmp_d[k])
                    else:
                        data[k] = [tmp_d[k]]

        for n in self.next: queue.append(n)

class BranchNode(Node):
    mermaid_style = NodeColorStyle.BranchNode
    mermaid_shape = NodeShape.BranchNode

    def post_init(self):
        self.passed_cond = set()
        if 'use_llm' not in self.conf: self.conf['use_llm'] = False
        if self.conf['use_llm']:
            tree = self.tree
            self.conf['prompt'] = tree.prompt_manager.prompts['branch_node_prompt']
            self.conf['return_json'] = True
            self.conf['format'] = {'item_id': str}

            if self.conf.get('backend', None):
                backend = self.conf['backend']
            elif (constructor := self.conf.get('backend_construct', None)):
                backend = constructor(tree.run_mode)
            else:
                backend = tree.llm_backend

            if tree.run_mode == 'mp':
                pipe = LLMPipe(
                        self.name,
                        llm=backend,
                        lock=tree.mp_lock,
                        run_time=tree.mp_manager.list(),
                        inout_log=tree.mp_manager.list(),
                        **self.conf
                        )
            else:
                pipe = LLMPipe(self.name, llm=backend, **self.conf)
            tree.pipe_manager[self.name] = pipe
            self.pipe = pipe

    def update(self, nodes):
        self.next = {}
        for cond, item in self.conf['next'].items():
            if (t := type(item)) is str:
                if item in nodes: self.next[cond] = [nodes[item]]
            elif t is list:
                self.next[cond] = []
                for name in item:
                    if name in nodes: self.next[cond].append(nodes[name])

    def get_mermaid(self, info=None):
        inps = ' & '.join(self.mermaid_inps)

        defines = self._get_mermaid_defines()

        t = info["detail"][self.name]["avg_time"] if info and self.name in info['detail'] else None
        if t is not None: t = f'|{t:.2f}s|'
        inline = self.mermaid_inline_passed if self.run_cnt else self.mermaid_inline
        inout_link = (None, None, inps, inline, t, self.name, None)

        links = [inout_link]
        for cond, arr in self.next.items():
            nexts = [n.name for n in arr]
            if 'exit' in nexts:
                t = f'|{cond}, total: {info["total_time"]:.2f}s|' if info else f'|{cond}|'
                links.append((None, None, self.name, self.mermaid_toexit_passed if t else self.mermaid_toexit, t, 'exit', None))
                nexts.remove('exit')

            if nexts:
                outline = self.mermaid_outline_passed if cond in self.passed_cond else self.mermaid_outline
                links.append((None, None, self.name, outline, f'|{cond}|', " & ".join(nexts), None))

        subg = []

        return defines, links, subg

    async def add_task(self, data, queue, dynamic_tasks):
        inps = await self.get_inps(queue)
        if self.conf['use_llm']:
            items = list(self.conf['next'].keys())
            items_text = '\n'.join([f'[#{i+1}] {t}' for i, t in enumerate(items)])

            retry = 0
            while retry < 5:
                if len(inps) == 1:
                    cond = await self.pipe.async_call(inps[0], items_text)
                else:
                    inps_t = '\n'.join(f'{k}: {v}' for k,v in zip(self.conf['inp'], inps))
                    cond = await self.pipe.async_call(inps_t, items_text)

                if cond and type(cond) is dict:
                    item_id = cond.get('item_id', '')
                    if len(item_id) > 1 and item_id[0] == '#' and item_id[1:].isdigit() and (i := int(item_id[1:]) - 1) < len(items):
                        cond = items[i]
                        break
                    elif len(item_id) and item_id.isdigit() and (i := int(item_id) - 1) < len(items):
                        cond = items[i]
                        break
                retry += 1
        elif 'code' in self.conf:
            if 'code_entry' in self.conf:
                exec(self.conf['code'])
                cond = locals()[self.conf['code_entry']](*inps)
            else:
                inps_dict = {k:v for k,v in zip(self.conf['inp'], inps)}
                cond = eval(self.conf['code'].format(**inps_dict))
        else:
            cond = inps[0]

        if type(cond) is dict: cond = str(cond)
        self.passed_cond.add(cond)
        if (nodes := self.next.get(cond, None)):
            for node in nodes:
                task = asyncio.create_task(node.run(data, queue, dynamic_tasks), name=node.name)
                dynamic_tasks.append(task)
            
            if self.run_cnt > 1:
                def q_del(q, k):
                    if k in q:
                        while not q[k].empty(): q[k].get_nowait()
                    if '_sub' in q:
                        for sub_q in q['_sub']: q_del(sub_q, k)

                sub_nodes = set()
                tmp = nodes[:]
                while tmp:
                    n = tmp.pop()
                    sub_nodes.add(n)
                    for nxt in n.next:
                        if nxt not in sub_nodes and type(nxt) is not BranchNode:
                            tmp.append(nxt)
              
                outs = set()
                for n in sub_nodes: outs.update(n.mermaid_outs)
                outs -= set(self.conf['inp'])
                for o in outs: q_del(queue, o)
                log.debug(f'[{self.name}] reset variables: {outs}')
        log.debug(f'[{self.name}] condition: {cond}, goto nodes: {nodes}')

    def current_mp_task(self, inps, data, queue, config=None):
        if self.conf['use_llm']:
            items = list(self.conf['next'].keys())
            items_text = '\n'.join([f'[#{i+1}] {t}' for i, t in enumerate(items)])

            retry = 0
            while retry < 5:
                if len(inps) == 1:
                    cond = self.pipe(inps[0], items_text)
                else:
                    inps_t = '\n'.join(f'{k}: {v}' for k,v in zip(self.conf['inp'], inps))
                    cond = self.pipe(inps_t, items_text)

                if cond and type(cond) is dict:
                    item_id = cond.get('item_id', '')
                    if len(item_id) > 1 and item_id[0] == '#' and item_id[1:].isdigit() and (i := int(item_id[1:]) - 1) < len(items):
                        cond = items[i]
                        break
                    elif len(item_id) and item_id.isdigit() and (i := int(item_id) - 1) < len(items):
                        cond = items[i]
                        break
                retry += 1
        elif 'code' in self.conf:
            if 'code_entry' in self.conf:
                exec(self.conf['code'])
                cond = locals()[self.conf['code_entry']](*inps)
            else:
                inps_dict = {k:v for k,v in zip(self.conf['inp'], inps)}
                cond = eval(self.conf['code'].format(**inps_dict))
        else:
            cond = inps[0]

        if type(cond) is dict: cond = str(cond)
        self.passed_cond.add(cond)
        if (nodes := self.next.get(cond, None)):
            for node in nodes: queue.put((node.name, config))
        log.debug(f'[{self.name}] condition: {cond}, goto nodes: {nodes}')

    def current_normal_task(self, inps, data, queue):
        if self.conf['use_llm']:
            items = list(self.conf['next'].keys())
            items_text = '\n'.join([f'[#{i+1}] {t}' for i, t in enumerate(items)])

            retry = 0
            while retry < 5:
                if len(inps) == 1:
                    cond = self.pipe(inps[0], items_text)
                else:
                    inps_t = '\n'.join(f'{k}: {v}' for k,v in zip(self.conf['inp'], inps))
                    cond = self.pipe(inps_t, items_text)

                if cond and type(cond) is dict:
                    item_id = cond.get('item_id', '')
                    if len(item_id) > 1 and item_id[0] == '#' and item_id[1:].isdigit() and (i := int(item_id[1:]) - 1) < len(items):
                        cond = items[i]
                        break
                    elif len(item_id) and item_id.isdigit() and (i := int(item_id) - 1) < len(items):
                        cond = items[i]
                        break
                retry += 1
        elif 'code' in self.conf:
            if 'code_entry' in self.conf:
                exec(self.conf['code'])
                cond = locals()[self.conf['code_entry']](*inps)
            else:
                inps_dict = {k:v for k,v in zip(self.conf['inp'], inps)}
                cond = eval(self.conf['code'].format(**inps_dict))
        else:
            cond = inps[0]

        if type(cond) is dict: cond = str(cond)
        self.passed_cond.add(cond)
        if (nodes := self.next.get(cond, None)): queue += nodes
        log.debug(f'[{self.name}] condition: {cond}, goto nodes: {nodes}')

    def export_as_comfyui(self):
        inps = {
            "text": ["TEXT"],
            "use_llm": ["BOOLEAN", {"default": self.conf.get('use_llm', False)}],
        }
        opt_inps = {}
        outs = list(self.conf['next'].keys())
        d = {
            "input": {
                "required": inps,
                "optional": opt_inps
            },
            "input_order": {"required": list(inps.keys())},
            "output": ["TEXT"] * len(outs),
            "output_is_list": [False] * len(outs),
            "output_name": outs,
            "name": self.name,
            "display_name": self.name,
            "description": f"{self.name} 分支流程",
            "python_module": "nodes",
            "category": "控制流",
            "output_node": False,
        }
        return {self.name: d}

    def __str__(self):
        if type(self.next) is dict:
            arr = [f'{cond} -> {[n.name for n in nodes]}' for cond, nodes in self.next.items()]
        else:
            arr = self.next
        return f"<{self.__class__.__name__}: {self.name}, next: {arr}>"

class CodeNode(Node):
    mermaid_style = NodeColorStyle.CodeNode
    mermaid_shape = NodeShape.CodeNode

    def _eval_format(self, item):
        if type(item) is str:
            return item.encode('unicode_escape').decode('utf-8')
        else:
            return item

    async def current_task(self, data, queue, dynamic_tasks):
        inps = await self.get_inps(queue)
        if 'code_entry' in self.conf:
            exec(self.conf['code'])
            out = locals()[self.conf['code_entry']](*inps)
        else:
            inps_dict = {k:self._eval_format(v) for k,v in zip(self.conf['inp'], inps)}
            out = eval(self.conf['code'].format(**inps_dict))
        self.set_out(out, data, queue)

    def current_mp_task(self, inps, data, queue, config=None):
        if 'code_entry' in self.conf:
            exec(self.conf['code'])
            out = locals()[self.conf['code_entry']](*inps)
        else:
            inps_dict = {k:self._eval_format(v) for k,v in zip(self.conf['inp'], inps)}
            out = eval(self.conf['code'].format(**inps_dict))
        self.set_out(out, data, config=config)
        for n in self.next: queue.put((n.name, config))
    
    def current_normal_task(self, inps, data, queue):
        if 'code_entry' in self.conf:
            exec(self.conf['code'])
            out = locals()[self.conf['code_entry']](*inps)
        else:
            inps_dict = {k:self._eval_format(v) for k,v in zip(self.conf['inp'], inps)}
            out = eval(self.conf['code'].format(**inps_dict))
        self.set_out(out, data)

class WebNode(Node):
    mermaid_style = NodeColorStyle.WebNode
    mermaid_shape = NodeShape.WebNode

    def post_init(self):
        tree = self.tree
        search_pipe = None
        if tree.is_async:
            if 'search_engine' in self.conf['web']:
                search_pipe = SearchPipe(self.name, **self.conf['web'])
            browser_pipe = BrowserPipe(self.name)
        else:
            if 'search_engine' in self.conf['web']:
                search_pipe = SearchPipe(
                        self.name,
                        lock=tree.mp_lock,
                        run_time=tree.mp_manager.list(),
                        inout_log=tree.mp_manager.list(),
                        **self.conf
                        )
            browser_pipe = BrowserPipe(self.name)

        tree.pipe_manager[self.name] = {'search': search_pipe, 'browser': browser_pipe}
        self.search_pipe = search_pipe
        self.browser_pipe = browser_pipe

    def current_mp_task(self, inps, data, queue, config=None):
        out = self.search_pipe(*inps)
        self.set_out(out, data, config=config)

        for n in self.next: queue.put((n.name, config))

    async def current_task(self, data, queue, dynamic_tasks):
        inps = await self.get_inps(queue)
        out = await self.search_pipe.async_call(*inps)

        browser_tasks = []
        for url in self.loop_nodes:
            task = asyncio.create_task(n.run(new_data, new_queue, loop_tasks))
            browser_tasks.append(task)
        while not all(t.done() for t in browser_tasks):
            await asyncio.gather(*browser_tasks)

        self.set_out(out, data, queue)

class BrowserNode(Node):
    ...

class ValueNode(Node):
    mermaid_style = NodeColorStyle.ValueNode
    mermaid_shape = NodeShape.ValueNode

    def _get_mermaid_defines(self):
        data_defs = [Data.mermaid_shape(d) for d in self.mermaid_data]
        n = 25
        if 'value' in self.conf:
            t = str(self.conf['value'])
        elif 'item' in self.conf:
            t = self.conf['item']
        if len(t) > n: t = t[:n] + '...'
        m = {'append': '\\+', 'assign': '='}[self.conf.get('mode', 'assign')]
        d = f'{self.name}\n{m} {t}'
        return [self.__class__.mermaid_shape(self.name, d)] + data_defs

    def get_inps_mp(self, data, config=None):
        def get_data(i):
            if i not in data: return DataState.VOID
            if type(data[i]) is list:
                if config and i in config['loop_index']:
                    loop_i = config['loop_index'][i]
                    return data[i][loop_i]
                elif DataState.VOID in data[i]:
                    return DataState.VOID
            return data[i]

        mode = self.conf.get('mode', 'assign')
        if mode == 'append':
            if (d := get_data(self.conf['out'])) is DataState.VOID: return []
            else: return [d]
        else:
            return [1]

    def current_mp_task(self, inps, data, queue, config=None):
        mode = self.conf.get('mode', 'assign')
        if mode == 'append':
            out = inps[0]
            out.append(self.conf['value'])
        else:
            out = self.conf['value']
        self.set_out(out, data, config=config)
        for n in self.next: queue.put((n.name, config))

    async def current_task(self, data, queue, dynamic_tasks):
        mode = self.conf.get('mode', 'assign')
        if 'value' in self.conf:
            v = self.conf['value']
        elif 'item' in self.conf:
            if 'inp' in self.conf: await self.get_inps(queue)
            if '->' in self.conf['item']:
                k1, k2 = self.conf['item'].split('->')
                v = data[data[k1][k2]]
            else:
                v = data[self.conf['item']]

        if mode == 'append':
            i = self.conf['out']
            out = await queue[i].get()
            out.append(v)
        else:
            out = v
        self.set_out(out, data, queue)

    def current_normal_task(self, inps, data, queue):
        mode = self.conf.get('mode', 'assign')
        if 'value' in self.conf:
            v = self.conf['value']
        elif 'item' in self.conf:
            if '->' in self.conf['item']:
                k1, k2 = self.conf['item'].split('->')
                v = data[data[k1][k2]]
            else:
                v = data[self.conf['item']]

        if mode == 'append':
            i = self.conf['out']
            out = data[i]
            out.append(v)
        else:
            out = v
        self.set_out(out, data)

class DatabaseNode(Node):
    def aa(self):
        ...

class ExitNode(Node):
    mermaid_style = NodeColorStyle.ExitNode
    mermaid_shape = NodeShape.ExitNode

    def get_mermaid(self, info=None):
        defines = [self.__class__.mermaid_shape(self.name)]
        return defines, [], []

    async def current_task(self, data, queue, dynamic_tasks):
        if self.conf:
            while not all(t.done() for t in dynamic_tasks if t.get_name() != 'exit'):
                await asyncio.sleep(0.1)

            self.reformat(data)

    def reformat(self, data):
        if self.conf and 'error_msg' not in data:
            ret = {}
            for k, v in self.conf.items():
                if type(v) is str:
                    d = data
                    for i in v.split('.'):
                        if i in d: d = d[i]
                        else: break
                    else:
                        ret[k] = d
                elif type(v) is list and type(v[0]) is dict:
                    d = []
                    for i in range(len(data[list(v[0].values())[0]])):
                        t = {}
                        for m, n in v[0].items():
                            if n in data:
                                t[m] = data[n][i]
                        if t: d.append(t)
                    if d: ret[k] = d

            for k in list(data.keys()): del data[k]
            for k in ret: data[k] = ret[k]

class FileNode(Node):
    mermaid_style = NodeColorStyle.FileNode
    mermaid_shape = NodeShape.FileNode

    def current_normal_task(self, inps, data, queue):
        info = self.conf.get('file_dir', None) or self.conf.get('file', None)
        if 'file' in self.conf:
            if (t := type(self.conf['file'])) is str:
                files = [Path(self.conf['file'])]
            elif t is list:
                files = [Path(f) for f in self.conf['file']]
        elif 'file_dir' in self.conf:
            files = [f for f in Path(self.conf['file_dir']).iterdir() if f.is_file()]

        md = []
        for file in files:
            if file.suffix == '.pdf':
                md_text = pymupdf4llm.to_markdown(file)
                md.append(md_text)
            else:
                with open(file, 'r') as f:
                    md.append(f.read())

        if 'file' in self.conf and t is str: md = md[0]

        self.set_out(md, data)
        log.debug(f'[{self.name}] read: {info} -> {self.conf["out"]}')
        

class PipeTree:
    def __init__(self, llm_backend, rag_backend, prompt_manager, name=None, pipeconf:dict=None, pipefile=None, run_mode='async'):
        self.name = name
        self.pipeconf = pipeconf
        self.pipefile = pipefile
        self.llm_backend = llm_backend
        self.rag_backend = rag_backend
        self.run_mode = run_mode
        self.is_async = run_mode=='async'
        self.perf = []
        if pipeconf is None and pipefile is not None: self.load(pipefile)

        self.start_nodes = []
        self.required_inputs = []
        self.exit_node = None
        self.prompt_manager = prompt_manager
        self.pipe_manager = {}
        self.node_manager = {}
        self.node_type = collections.defaultdict(set)
        if run_mode == 'mp':
            self.mp_manager = mp.Manager()
            self.mp_lock = self.mp_manager.Lock()

        self._init()
        self._check()

    def load(self, pipefile):
        if type(pipefile) is str: pipefile = Path(pipefile)
        m = importlib.import_module(pipefile.stem)
        self.pipeconf = m.pipeline
        if self.name is None: self.name = pipefile.stem

    def _check(self):
        node_names = set(self.node_manager.keys())
        data = set()
        for n in self.node_manager.values(): data.update(n.mermaid_data)
        if (conflict_names := node_names & data):
            log.error(f'conflict between the pipe name and the data name, which will cause errors when drawing mermaid flowchart: {conflict_names}')
            exit()

    def _find_start_nodes(self):
        conf = self.pipeconf
        deps = set(['exit'])
        for p in conf:
            if (nt := conf[p].get('next', None)):
                if type(nt) is dict:
                    arr = []
                    for v in nt.values():
                        if type(v) is str: arr.append(v)
                        elif type(v) is list: arr += v
                    deps.update(arr)
                elif type(nt) is list:
                    deps.update(nt)
            deps.update(conf[p].get('pipe_in_loop', []))

        self.start_nodes = [self.node_manager[i] for i in list(set(conf.keys()) - deps)]

    def _init(self):
        if 'exit' not in self.pipeconf: self.pipeconf['exit'] = {}
        for name, conf in self.pipeconf.items():
            if name == 'exit':
                self.exit_node = node = ExitNode(name, conf, self)
            elif 'rag_param' in conf:
                node = RAGNode(name, conf, self)
            elif 'pipe_in_loop' in conf:
                node = LoopNode(name, conf, self)
            elif 'use_llm' in conf or type(conf.get('next', None)) is dict:
                node = BranchNode(name, conf, self)
            elif 'code' in conf:
                node = CodeNode(name, conf, self)
            elif 'web' in conf:
                node = WebNode(name, conf, self)
            elif 'value' in conf or 'item' in conf:
                node = ValueNode(name, conf, self)
            elif 'prompt' in conf:
                node = LLMNode(name, conf, self)
            elif 'file' in conf or 'file_dir' in conf:
                node = FileNode(name, conf, self)
            else:
                log.error(f"Unable to identify node type: [{name}] {conf}")
                exit()
            self.node_manager[name] = node

        all_outs = set()
        for n in self.node_manager.values():
            n.update(self.node_manager)
            self.node_type[n.__class__].add(n.name)
            self.node_type[Data].update(n.mermaid_data)
            all_outs.update(n.mermaid_outs)

        self.required_inputs = self.node_type[Data] - all_outs
        self.node_type[InputData] = self.required_inputs

        self._find_start_nodes()
        if not self.start_nodes:
            log.error(f"Can't find start entry in pipes.")
            exit()
        log.debug(f"'{self.name}' tree initialization successful, start nodes: {self.start_nodes}")

    def export_conf(self):
        conf = copy.deepcopy(self.pipeconf)
        for pipe_conf in conf.values():
            if 'format' in pipe_conf:
                for k, v in pipe_conf['format'].items():
                    pipe_conf['format'][k] = str(v)
        return conf

    def reset(self):
        self.perf = []
        for node in self.node_manager.values(): node.reset()

    def tree2mermaid(self, info=None):
        mermaid = 'graph TD'
        indent = ' '*4
        defines = set()
        links = set()
        subgraphs = set()
        passed_nodes = set()
        nodes = self.start_nodes[:]
        while nodes:
            n = nodes.pop(0)
            passed_nodes.add(n)
            defs, lks, subg = n.get_mermaid(info)
            defines.update(defs)
            links.update(lks)
            subgraphs.update(subg)
            for i in n.loop_nodes + (n.next if type(n.next) is list else reduce(lambda a,b:a+b, n.next.values())):
                if i not in passed_nodes: nodes.append(i)

        links_d = {}
        for link in links:
            item = (link[0], link[2], link[5])
            if item in links_d:
                if link[4]: links_d[item] = link
            else:
                links_d[item] = link
        links_str = []
        links_style = []

        for inps, inline, self.name, outline, t, outs, link_style in links_d.values():
            if t is None: t = ''
            if inps:
                link = f'{inps} {inline} {self.name} {outline}{t} {outs}'
            else:
                link = f'{self.name} {outline}{t} {outs}'
            if link_style is not None:
                links_style.insert(0, link_style)
                links_str.insert(0, link)
            else:
                links_str.append(link)

        mermaid += f'\n{indent}%% ========================'
        mermaid += f'\n{indent}%% Nodes definition section'
        mermaid += f'\n{indent}%% ========================\n'
        for i in defines: mermaid += f'{indent}{i}\n'

        mermaid += f'\n{indent}%% ========================'
        mermaid += f'\n{indent}%% Links definition section'
        mermaid += f'\n{indent}%% ========================\n'
        for i in links_str: mermaid += f'{indent}{i}\n'

        mermaid += f'\n{indent}%% ================'
        mermaid += f'\n{indent}%% Subgraph section'
        mermaid += f'\n{indent}%% ================\n'
        for i in subgraphs:
            sub_items = "\n".join([indent*2+j for j in i[1:]])
            mermaid += f'{indent}subgraph {i[0]}\n{sub_items}\n{indent}end\n'

        mermaid += f'\n{indent}%% ========================'
        mermaid += f'\n{indent}%% Style definition section'
        mermaid += f'\n{indent}%% ========================\n'
        for c, items in self.node_type.items():
            name = c.__name__.upper()
            mermaid += f'{indent}classDef {name} {c.mermaid_style}\n'
            if items: mermaid += f'{indent}class {",".join(items)} {name}\n'
        for i, style in enumerate(links_style):
            mermaid += f'{indent}linkStyle {i} {style}\n'

        return mermaid

    def perf2mermaid(self):
        mermaid = 'gantt\ntitle Task Timeline\ndateFormat  x\naxisFormat  %M:%S.%L\n'
        base_time = self.perf[0][2]
        data = collections.defaultdict(list)
        loop_end = {}
        for name, e, start_time, end_time in self.perf:
            if 'pid' in name:
                arr = name.split(': ')
                n = f'{arr[0]}_{int(arr[1]):0>2d}'
            else:
                n = name
            s = (start_time - base_time) * 1000
            Δ = (end_time - start_time) * 1000
            data[n].append((e, s, Δ))
            # if e in pipe:
            #     if pipe[e]['mode'] == 'loop':
            #         loop_end[f'{e}_end'] = (n, e, s)
            #     else:
            #         data[n].append((e, s, Δ))
            # elif e in loop_end:
            #     n, e, s_ = loop_end[e]
            #     Δ_ = s + Δ - s_
            #     data[n].append((e, s_, Δ_))
        
        for k in sorted(data.keys()):
            mermaid += f'section {k}\n'
            for e, s, Δ in data[k]:
                # if pipe[e]['mode'] == 'loop':
                #     mermaid += f'{e}: done, {s:.0f}, {Δ:.0f}ms\n'
                # else:
                mermaid += f'{e}: {s:.0f}, {Δ:.0f}ms\n'

        return mermaid

    async def async_run(self, inp_data):
        self.reset()
        data = copy.deepcopy(inp_data)
        dynamic_tasks = []
        queue = collections.defaultdict(asyncio.Queue)
        for k, v in data.items(): queue[k].put_nowait(v)
        try:
            await asyncio.gather(*[asyncio.create_task(n.run(data, queue, dynamic_tasks)) for n in self.start_nodes])
            while not all(t.done() for t in dynamic_tasks):
                await asyncio.gather(*dynamic_tasks)
        except Exception as e:
            error_msg = traceback.format_exc()
            data['error_msg'] = error_msg
            log.error(f'[{self.name}]:\n{error_msg}')
        return data

    def mp_task(self, pid, data, task_queue, perf_queue):
        name = f'pid: {pid}'
        lock = self.mp_lock
        log.debug(f'{name}, start')

        try:
            while True:
                try:
                    node_name, config = task_queue.get_nowait()
                    if node_name == 'exit':
                        # if task_queue.qsize() == 0:
                        task_queue.put(('exit', None))
                        log.debug(f'{name}, exit')
                        break
                except queue.Empty: continue

                self.node_manager[node_name].run(name, data, task_queue, perf_queue, config)
        except Exception as e:
            error_msg = traceback.format_exc()
            with lock: data['error_msg'] = error_msg
            log.error(f'[{name}]:\n{error_msg}')
            task_queue.put(('exit', {}))
            log.debug(f'{name}, exit')

    def mp_run(self, inp_data, core_num=4):
        self.reset()
        task_queue = self.mp_manager.Queue()
        perf_queue = self.mp_manager.Queue()
        data       = self.mp_manager.dict()
        for n in self.start_nodes: task_queue.put((n.name, None))
        for k, v in inp_data.items(): data[k] = v

        processes = []
        for i in range(core_num):
            p = mp.Process(target=self.mp_task, args=(i, data, task_queue, perf_queue))
            processes.append(p)
            p.start()
        for p in processes: p.join()
        while not perf_queue.empty(): self.perf.append(perf_queue.get())

        data = dict(data)
        self.exit_node.reformat(data)
        return data

    def normal_run(self, inp_data):
        self.reset()
        data = copy.deepcopy(inp_data)

        try:
            queue = self.start_nodes[:]
            while queue:
                node = queue.pop(0)
                node.run(data, queue)
        except Exception as e:
            error_msg = traceback.format_exc()
            data['error_msg'] = error_msg
            log.error(f'[{self.name}]:\n{error_msg}')
        return data

    async def server_run(self, inp_data):
        self.reset()
        data = copy.deepcopy(inp_data)

        queue = self.start_nodes[:]
        while queue:
            node = queue.pop(0)
            pre_data = copy.deepcopy(data)
            node.run(data, queue)

            arr = []
            for k in data:
                if k not in pre_data or data[k] != pre_data[k]:
                    arr.append(k)
            if arr:
                yield {node.name: {k: data[k] for k in arr}}

            await asyncio.sleep(0)

    def __str__(self):
        loop_num = sum([type(n) is LoopNode for n in self.node_manager.values()])
        branch_num = sum([type(n) is BranchNode for n in self.node_manager.values()])
        return f"<PipeTree '{self.name}': nodes: {len(self.node_manager)}, loop: {loop_num}, branch: {branch_num}, mode: {self.run_mode}>"

    def __repr__(self):
        return self.__str__()



from pydantic import BaseModel
from fastapi.staticfiles import StaticFiles
from starlette.websockets import WebSocketState
from fastapi import FastAPI, HTTPException, WebSocket, WebSocketDisconnect
from fastapi.responses import HTMLResponse, FileResponse, RedirectResponse

import time
import copy
import uuid
import json
import heapq
import struct
import asyncio
import threading
from io import BytesIO
from pathlib import Path
from itertools import chain
from functools import partial
from PIL import Image, ImageOps, PngImagePlugin
from typing import Optional, Dict, Any, Callable
from .log import log

PipelineServer = server = FastAPI(title='llmpipeline server')
PipelineServer.api = None
web_dir = 'web'
web_root = Path(f'{__file__[:__file__.rindex("/")]}/{web_dir}')
server.mount("/web", StaticFiles(directory=web_root, html=True), name="llmpipeline server web")

class BinaryEventTypes:
    PREVIEW_IMAGE = 1
    UNENCODED_PREVIEW_IMAGE = 2

class WSConnectionManager:
    def __init__(self):
        self.active_connections: Dict[str, WebSocket] = {}

    async def connect(self, websocket: WebSocket, sid: str):
        await websocket.accept()
        self.active_connections[sid] = websocket

    def disconnect(self, sid: str):
        if sid in self.active_connections:
            del self.active_connections[sid]

    async def send(self, event, data, sid=None):
        if event == BinaryEventTypes.UNENCODED_PREVIEW_IMAGE:
            await self.send_image(data, sid=sid)
        elif isinstance(data, (bytes, bytearray)):
            await self.send_bytes(event, data, sid)
        else:
            message = {"type": event, "data": data}
            await self.send_json(message, sid)

    async def send_json(self, message: dict, sid: Optional[str] = None):
        if sid:
            websocket = self.active_connections.get(sid)
            if websocket and websocket.application_state == WebSocketState.CONNECTED:
                await websocket.send_json(message)
        else:
            for ws in self.active_connections.values():
                if ws.application_state == WebSocketState.CONNECTED:
                    await ws.send_json(message)

    async def send_bytes(self, data: bytes, sid: Optional[str] = None):
        if sid:
            websocket = self.active_connections.get(sid)
            if websocket and websocket.application_state == WebSocketState.CONNECTED:
                await websocket.send_bytes(data)
        else:
            for ws in self.active_connections.values():
                if ws.application_state == WebSocketState.CONNECTED:
                    await ws.send_bytes(data)

    async def send_image(self, image_data, sid: Optional[str] = None):
        image_type = image_data[0]
        image = image_data[1]
        max_size = image_data[2]
        if max_size is not None:
            resampling = Image.Resampling.BILINEAR if hasattr(Image, 'Resampling') else Image.ANTIALIAS
            image = ImageOps.contain(image, (max_size, max_size), resampling)
        type_num = 1 if image_type == "JPEG" else 2
        bytes_io = BytesIO()
        header = struct.pack(">I", type_num)
        bytes_io.write(header)
        image.save(bytes_io, format=image_type, quality=95, compress_level=1)
        preview_bytes = bytes_io.getvalue()
        await self.send_bytes(BinaryEventTypes.PREVIEW_IMAGE, preview_bytes, sid=sid)

class TaskQueue:
    def __init__(self, loop=None, max_history_size=10000):
        self.loop = loop
        self.max_history_size = max_history_size
        self.mutex = threading.RLock()
        self.not_empty = threading.Condition(self.mutex)
        self.task_counter = 0
        self.queue = []
        self.currently_running = {}
        self.history = {}
        self.flags = {}

    def put(self, item):
        with self.mutex:
            heapq.heappush(self.queue, item)
            self.queue_updated_broadcast()
            self.not_empty.notify()

    def get(self, timeout=None):
        with self.not_empty:
            while len(self.queue) == 0:
                self.not_empty.wait(timeout=timeout)
                if timeout is not None and len(self.queue) == 0:
                    return None
            item = heapq.heappop(self.queue)
            task_id = self.task_counter
            self.currently_running[task_id] = copy.deepcopy(item)
            self.task_counter += 1
            self.queue_updated_broadcast()
            return task_id, item

    def task_done(self, task_id, history_result, status):
        with self.mutex:
            prompt = self.currently_running.pop(task_id)
            if len(self.history) > self.max_history_size:
                self.history.pop(next(iter(self.history)))

            self.history[prompt[1]] = {
                "prompt": prompt,
                'status': status,
            } | history_result
            self.queue_updated_broadcast()

    def queue_updated_broadcast(self):
        event = "status"
        data = {"status": self.get_queue_info()}
        self.loop.call_soon_threadsafe(ws_msges.put_nowait, (event, data, None))

    def get_queue_info(self):
        prompt_info = {
            'exec_info': {
                'queue_remaining': self.get_tasks_remaining()
            }
        }
        return prompt_info

    def get_current_queue(self):
        with self.mutex:
            out = []
            for x in self.currently_running.values():
                out += [x]
            return (out, copy.deepcopy(self.queue))

    def get_tasks_remaining(self):
        with self.mutex:
            return len(self.queue) + len(self.currently_running)

    def wipe_queue(self):
        with self.mutex:
            self.queue = []
            self.server.queue_updated()

    def delete_queue_item(self, function):
        with self.mutex:
            for x in range(len(self.queue)):
                if function(self.queue[x]):
                    if len(self.queue) == 1:
                        self.wipe_queue()
                    else:
                        self.queue.pop(x)
                        heapq.heapify(self.queue)
                    self.server.queue_updated()
                    return True
        return False

    def get_history(self, prompt_id=None, max_items=None, offset=-1):
        with self.mutex:
            if prompt_id is None:
                out = {}
                i = 0
                if offset < 0 and max_items is not None:
                    offset = len(self.history) - max_items
                for k in self.history:
                    if i >= offset:
                        out[k] = self.history[k]
                        if max_items is not None and len(out) >= max_items:
                            break
                    i += 1
                return out
            elif prompt_id in self.history:
                return {prompt_id: copy.deepcopy(self.history[prompt_id])}
            else:
                return {}

    def wipe_history(self):
        with self.mutex:
            self.history = {}

    def delete_history_item(self, id_to_delete):
        with self.mutex:
            self.history.pop(id_to_delete, None)

    def set_flag(self, name, data):
        with self.mutex:
            self.flags[name] = data
            self.not_empty.notify()

    def get_flags(self, reset=True):
        with self.mutex:
            if reset:
                ret = self.flags
                self.flags = {}
                return ret
            else:
                return self.flags.copy()

class TaskWorker(threading.Thread):
    def __init__(self, queue=None, loop=None):
        name = self.__class__.__name__
        threading.Thread.__init__(self, name=name, daemon=True)
        self.queue = queue
        self.loop = loop
        log.debug(f'{name} thread start')

    def run(self):
        name = threading.current_thread().name

        while True:
            queue_item = self.queue.get(timeout=1000)
            if queue_item is not None:
                task_id, (_, prompt_id, task_data, extra_data) = queue_item
                log.debug(f'{name}:\ntask_id: {task_id}\nprompt_id: {prompt_id}\ntask_data: {task_data}\nextra_data: {extra_data}')

                out = self.run_task(prompt_id, task_data, extra_data)
                his = {
                    "outputs": {
                        out['node']: out['output'],
                    },
                    "meta": {},
                }

                self.queue.task_done(
                    task_id,
                    his,
                    status={
                        "status_str": 'success',
                        "completed": True,
                        "messages": None,
                    }
                )

                self.send_msg("executing", {"node": None, "prompt_id": prompt_id}, extra_data['client_id'])

    def run_task(self, prompt_id, task_data, extra_data):
        ws_id = extra_data.get('client_id', None)
        self.send_msg("execution_start", {"prompt_id": prompt_id}, ws_id)
        self.send_msg("execution_cached", {"nodes":[], "prompt_id": prompt_id}, ws_id)

        out = None
        if PipelineServer.api:
            pipe = PipelineServer.api.pipeline_manager.add_pipe('comfyUI', is_seq=True)
            out = pipe.run(prompt_id, ws_id, task_data, self.send_msg)

        return out

    def send_msg(self, event, data, ws_id=None):
        data |= {"timestamp": int(time.time() * 1000)}
        self.loop.call_soon_threadsafe(ws_msges.put_nowait, (event, data, ws_id))

ws_msges = asyncio.Queue()
ws_manager = WSConnectionManager()
task_queue = TaskQueue()

async def ws_loop():
    while True:
        msg = await ws_msges.get()
        log.debug(f'ws: {msg}')
        await ws_manager.send(*msg)

@server.websocket("/ws")
async def websocket_endpoint(ws: WebSocket, clientId: Optional[str] = None):    
    # init process
    if task_queue.loop is None:
        loop = asyncio.get_running_loop()
        task_queue.loop = loop
        TaskWorker(queue=task_queue, loop=loop).start()
        asyncio.create_task(ws_loop())

    sid = clientId or uuid.uuid4().hex
    await ws_manager.connect(ws, sid)
    try:
        d = {
            'exec_info': {
                'queue_remaining': 0,
            },
        }
        await ws_manager.send("status", {"status": d, 'sid': sid}, sid)

        # if prompt_server.client_id == sid and prompt_server.last_node_id is not None:
        #     await prompt_server.send_json("executing", {"node": prompt_server.last_node_id}, sid)

        while True:
            data = await ws.receive()
            if data["type"] == "websocket.disconnect":
                break
    except WebSocketDisconnect:
        pass
    finally:
        ws_manager.disconnect(sid)

@server.get("/extensions")
async def get_extensions():
    c = chain(
        web_root.glob('extensions/core/*.js'),
        web_root.glob('extensions/*/lib/*.js')
    )
    extensions = list(map(lambda x: f'/{web_dir}/{x.relative_to(web_root)}', c))
    return extensions


@server.get("/object_info")
async def get_object_info():
    with open(web_root / 'addition/my_nodes.json', 'r') as f:
        out = json.load(f)
    with open(web_root / 'addition/nodes.json', 'r') as f:
        out |= json.load(f)
    if server.api: out |= server.api.pipeline_manager.export_nodes()
    return out

@server.get("/prompt")
async def get_prompt():
    return task_queue.get_queue_info()

class PromptData(BaseModel):
    client_id: str | None = None
    prompt: dict
    extra_data: dict | None = None

@server.post("/prompt")
async def post_prompt(data: PromptData):
    if data.prompt:
        extra_data = data.extra_data or {}
        if data.client_id: extra_data["client_id"] = data.client_id

        prompt_id = str(uuid.uuid4())
        task_queue.put((0, prompt_id, data.prompt, extra_data))

        ret = {
            "prompt_id": prompt_id,
            "number": 0, 
            "node_errors": []
        }
        return ret
    else:
        raise HTTPException(status_code=400, detail={"error": "no prompt", "node_errors": []})

@server.get("/queue")
async def get_queue():
    running, pending = task_queue.get_current_queue()
    queue_info = {
        'queue_running': running,
        'queue_pending': pending,
    }
    return queue_info

@server.get("/view_model/{item}")
async def view_model(item: str, filename: Optional[str] = None):
    if not filename:
        raise HTTPException(status_code=404, detail="Item not found")

    if item == 'model':
        ...

    image_path = web_root / 'images/ChatGPT.png'
    return FileResponse(image_path)

@server.get("/view")
async def view_image(filename: Optional[str] = None, type: Optional[str] = None, subfolder: Optional[str] = None):
    if not filename:
        raise HTTPException(status_code=404, detail="Item not found")

    if filename not in ['doubao.png', 'ChatGPT.png', 'grok.png', 'wenxin.png']:
        image_path = web_root / 'images/ChatGPT.png'
    else:
        image_path = web_root / f'images/{filename}'
    return FileResponse(image_path)



import unittest
from unittest.mock import patch, MagicMock
from pathlib import Path
from sigmaflow.manager import PromptManager

# filepath: gits/SigmaFlow/sigmaflow/test_manager.py

class TestPromptManager(unittest.TestCase):
    @patch('sigmaflow.manager.log')
    @patch('sigmaflow.manager.Prompt')
    @patch('sigmaflow.manager.importpath')
    def test_initialization_with_default_dir(self, mock_importpath, mock_prompt, mock_log):
        pm = PromptManager()
        self.assertEqual(pm.prompts_dir, Path('.'))
        self.assertTrue(mock_log.debug.called)

    @patch('sigmaflow.manager.log')
    @patch('sigmaflow.manager.Prompt')
    @patch('sigmaflow.manager.importpath')
    def test_initialization_with_custom_dir(self, mock_importpath, mock_prompt, mock_log):
        custom_dir = '/custom/prompts'
        pm = PromptManager(prompts_dir=custom_dir)
        self.assertEqual(pm.prompts_dir, Path(custom_dir))
        self.assertTrue(mock_log.debug.called)

    @patch('sigmaflow.manager.log')
    @patch('sigmaflow.manager.Prompt')
    @patch('sigmaflow.manager.importpath')
    def test_load_prompts(self, mock_importpath, mock_prompt, mock_log):
        mock_importpath.return_value = MagicMock(prompt="test_prompt", keys=["key1", "key2"])
        mock_prompt.return_value = MagicMock()

        pm = PromptManager()
        pm.buildin_prompts_dir = Path('/mock/build_in_prompts')
        pm.prompts_dir = Path('/mock/prompts')

        with patch('pathlib.Path.glob', return_value=[Path('mock_prompt.py')]):
            pm.load_prompts()

        self.assertIn('mock_prompt', pm.prompts)
        self.assertTrue(mock_log.debug.called)

    @patch('sigmaflow.manager.log')
    @patch('sigmaflow.manager.Prompt')
    def test_get_prompt_by_name(self, mock_prompt, mock_log):
        mock_prompt_instance = MagicMock()
        pm = PromptManager()
        pm.prompts['test_prompt'] = mock_prompt_instance

        result = pm.get('test_prompt')
        self.assertEqual(result, mock_prompt_instance)

    @patch('sigmaflow.manager.log')
    @patch('sigmaflow.manager.Prompt')
    def test_get_prompt_by_dict(self, mock_prompt, mock_log):
        mock_prompt_instance = MagicMock()
        mock_prompt.return_value = mock_prompt_instance

        pm = PromptManager()
        prompt_dict = {'prompt': 'test', 'keys': ['key1']}
        result = pm.get(prompt_dict)

        self.assertEqual(result, mock_prompt_instance)
        self.assertIn('prompt_0', pm.prompts)

    @patch('sigmaflow.manager.log')
    def test_get_prompt_not_found(self, mock_log):
        pm = PromptManager()
        with self.assertRaises(KeyError):
            pm.get('non_existent_prompt')

    @patch('sigmaflow.manager.log')
    @patch('sigmaflow.manager.Prompt')
    def test_get_prompt_duplicate_name(self, mock_prompt, mock_log):
        pm = PromptManager()
        pm.prompts['prompt_0'] = MagicMock()

        prompt_dict = {'prompt': 'test', 'keys': ['key1']}
        with self.assertRaises(AssertionError):
            pm.get(prompt_dict)

if __name__ == '__main__':
    unittest.main()


from openai import OpenAI
from openai import AsyncOpenAI
from . import *

def llm_client(is_async=False):
    if is_async:
        return async_completion
    else:
        return completion

def completion(text):
    client = OpenAI(
        api_key=os.getenv("OPENAI_API_KEY"),
        base_url=os.getenv("OPENAI_API_URL"),
        max_retries=20,
        # timeout=6000.0,
    )

    chat_completion = client.chat.completions.create(
        messages=[
            {
                "role": "user",
                "content": text,
            }
        ],
        model=os.getenv("OPENAI_API_MODEL")
    )

    return chat_completion.choices[0].message.content

async def async_completion(text):
    client = AsyncOpenAI(
        api_key=os.getenv("OPENAI_API_KEY"),
        base_url=os.getenv("OPENAI_API_URL"),
        max_retries=20,
        # timeout=6000.0,
    )
    if (t := type(text)) is str:
        msg = [
            {
                "role": "user",
                "content": text,
            }
        ]
    elif t is list:
        msg = []
        for role, c in zip(['user', 'assistant']*len(text), text):
            msg.append({
                "role": role,
                "content": c,
            })

    async with llm_sem:
        chat_completion = await client.chat.completions.create(
            messages=msg,
            model=os.getenv("OPENAI_API_MODEL")
        )

    return chat_completion.choices[0].message.content



import setuptools
import requests
from functools import reduce
import xml.etree.ElementTree as ET
import re

pkg_name = 'SigmaFlow'

def get_new_version():
    pkg_url = f'https://pypi.org/rss/project/{pkg_name}/releases.xml'
    root = ET.fromstring(requests.get(pkg_url).text)
    old_version = root[0].findall('item')[0].find('title').text
    num = reduce(lambda a,b:a*100+b, map(int, old_version.split('.')))
    num += 1
    arr = []
    while num:
        arr.append(num % 100)
        num //= 100
    arr += [0] * max(0, (3-len(arr)))
    version = '.'.join(map(str, arr[::-1]))

    return version

def parse_requirements(filename):
    with open(filename, "r") as f:
        lines = f.read().splitlines()

    requires = []

    for line in lines:
        if "http" in line:
            pkg_name_without_url = line.split('@')[0].strip()
            requires.append(pkg_name_without_url)
        else:
            requires.append(line)

    return requires

with open("README.md", "r") as f:
    long_description = f.read()
    m = re.findall( r"```mermaid.*?```", long_description, flags=re.DOTALL)
    long_description = long_description.replace(m[0], '![pipe demo](https://raw.githubusercontent.com/maokangkun/SigmaFlow/main/assets/demo_pipe.png)').replace(m[1], '![perf demo](https://raw.githubusercontent.com/maokangkun/SigmaFlow/main/assets/demo_perf.png)')

setuptools.setup(
    name=pkg_name.lower(),  # Replace with your own username
    version=get_new_version(),
    author="maokangkun",
    author_email="maokangkun@pjlab.prg.cn",
    description=f"{pkg_name} is a Python package designed to optimize the performance of task-flow related to LLMs or MLLMs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=f"https://github.com/maokangkun/{pkg_name}",
    packages=setuptools.find_packages(),
    install_requires=parse_requirements("requirements.txt"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    zip_safe=False,
    entry_points = {
        'console_scripts': ['sigmaflow=sigmaflow.cmd:main'],
    },
    # scripts=['bin/funniest-joke']
)



import os
import json
import httpx
import requests

def search(query, mkt='zh-CN', count=5):
    subscription_key = os.environ['BING_SEARCH_V7_SUBSCRIPTION_KEY']
    endpoint = os.environ['BING_SEARCH_V7_ENDPOINT']
    params = {'q': query, 'mkt': mkt, 'count': count}
    headers = {'Ocp-Apim-Subscription-Key': subscription_key}
    response = requests.get(endpoint, headers=headers, params=params)
    response.raise_for_status()
    res = response.json()
    data = []
    for r in res['webPages']['value']:
        d = {
            'title': r['name'],
            'url': r['url'],
            'snippet': r['snippet'],
            'date': r['dateLastCrawled'],
        }
        data.append(d)
    return data

async def async_search(query, mkt='zh-CN', count=5):
    subscription_key = os.environ['BING_SEARCH_V7_SUBSCRIPTION_KEY']
    endpoint = os.environ['BING_SEARCH_V7_ENDPOINT']
    params = {'q': query, 'mkt': mkt, 'count': count}
    headers = {'Ocp-Apim-Subscription-Key': subscription_key}

    async with httpx.AsyncClient(verify=False) as client:
        async with client.stream(
            "GET",
            endpoint,
            headers=headers,
            params=params,
            timeout=600,
        ) as stream:
            # error
            if stream.status_code != 200:
                raise HTTPException(status_code=500, detail=f'Bing API Error! url: {endpoint}, data: {params}')

            content = (await stream.aread()).decode("utf8")
            res = json.loads(content)
            data = []
            for r in res['webPages']['value']:
                d = {
                    'title': r['name'],
                    'url': r['url'],
                    'snippet': r['snippet'],
                    'date': r['dateLastCrawled'],
                }
                data.append(d)
            return data
